import contextlib, io  # to be able to capture stdout entry, so what is print(...) by a module
import time
from datetime import datetime  # time
import os  # to get filename from path
import numpy as np  # for matrix math with numpy
import logging
import dill  # to save and load a class object in a binary file

from pyrftl.func.various_stuff import opm_dist_idx1_idx2
logger = logging.getLogger('PyRFTL')


def create_save_folder(save_folder_path, analysis_name=None, t0=None) :
    # if no name for the current analysis, create one
    if analysis_name is None :
        if t0 is None :
            t0 = time.time()
        analysis_name = 'PyRFTL_' + datetime.fromtimestamp(t0).strftime("%Y_%m_%d_%Hh%M")

    # check if the parent save directory exist
    if not os.path.exists(save_folder_path) :
        print('Save folder does not exist')
        raise ValueError ('No folder at path : ' + str(save_folder_path))

    # check if the directory name for the current analysis is free, change the analysis name if not
    i = 0
    analysis_name_tmp = analysis_name
    while os.path.exists(os.path.join(save_folder_path, analysis_name_tmp)) :
        analysis_name_tmp = analysis_name+'_'+str(i)
        i = i+1

    # create the analysis directory
    os.mkdir(os.path.join(save_folder_path, analysis_name_tmp))
    logger.info('create folder : ' + str(os.path.join(save_folder_path, analysis_name_tmp)))

    return os.path.join(save_folder_path, analysis_name_tmp), analysis_name_tmp, t0


def data_list_to_str(list_data, data_separator=';', end_separator='\n', decimal_separator=None):
    # take a list of data in input, and create a str with them :
    # data[0] data_separator data[1] data_separator ... data[-1] end_separator
    output = ''

    for data in list_data:
        try:
            if data is not None:
                str_data = str(data)

                if decimal_separator is not None and decimal_separator != '.' :
                    try :
                        float(data)  # test if we have a decimal number (float, np.float, ...)
                        # if not a decimal, we will have an error so exit the try statement
                        str_data = str_data.replace('.', decimal_separator)
                    except :
                        pass  # do nothing, not a float, no need to change the decimal separator

                output = output + str_data + data_separator
            else:
                output = output + '' + data_separator
        except:
            output = output + '' + data_separator

    output = output + end_separator
    return output


def save_results_csv(param, pair_list, pair_first_rms, pair_second_rms=None, index_min_rms_tolerance=[]):
    logger.info('saving results in csv ...')
    print('saving results in csv ...')

    # create the file
    file_path = os.path.join(param['save_folder_path'], param['analysis_name'])
    try:
        file = open(file_path + ".csv", "x")
    except:  # this should never happen due to precedent code, but if it is, we want to be sure we can create a new file
        i = 0
        while i < 100:
            i = i + 1
            try:
                file = open(file_path + "_" + str(i) + ".csv", "x")
                file_path = file_path + "_" + str(i)
                break
            except:
                pass

    sep_row = param['output_csv']['output_csv_row_separator']
    sep_column = param['output_csv']['output_csv_column_separator']
    sep_dec = param['output_csv']['output_csv_decimal_separator']

    ######## HEADER ########
    # header
    file.write("Analysis of lens pairs for optical remote focusing." + sep_row)
    file.write(sep_row)
    file.write("Analysis name : " + param['analysis_name'] + sep_row)
    file.write("File save time : " + str(datetime.now()) + sep_row)
    file.write("All distances are in mm (except wavelenghts : in nm), angles are in degrees,"
               " RMS wavefront error is in waves" + sep_row)
    file.write(sep_row)
    file.write("parameters :" + sep_row)
    for pa in param:
        if pa == 'save_folder_path' and param['output_csv']['hide_path_in_csv']:
            file.write(str(pa) + sep_column + 'hidden' + sep_row)
        elif pa == 'lenses_param':
            lenses_param_tmp = param[pa].copy()
            lenses_param_tmp['LensesFolderPath'] = 'hidden'
            file.write(str(pa) + sep_column + str(lenses_param_tmp) + sep_row)
        else:
            file.write(str(pa) + sep_column + str(param[pa]) + sep_row)

    file.write(sep_row)
    file.write("WARNING : if your lenses files don't take into account the reduction of lens diameter due to their"
               " mounting, the cut-off angle can be overestimated !" + sep_row)
    file.write("WARNING : Coating is not taken into account, please check if your lenses are designed for the "
               "right wavelengths !" + sep_row)
    file.write(sep_row)

    ######## DATA ########
    # data
    file.write("pairs :" + sep_row)
    file.write(data_list_to_str(['id', 'lens1 filename', 'lens2 filename', 'config', 'pair short name',
                                 'lens1 id', 'lens2 id',
                                 'lens1 EFL', 'lens2 EFL',
                                 'first analysis RMS wavefront error',
                                 'second analysis RMS wavefront error', '<lambda/14', 'cut-off angle', 'ffl',
                                 'distance between lenses',
                                 'bfl', 'optimum bfl for RMS wavefront', 'efl', "distance F to F'",
                                 "distance first to last surface of pair lenses",
                                 "lens1 total path", "lens2 total path", 'lens1 min diameter', 'lens1 max diameter',
                                 'lens2 min diameter', 'lens2 max diameter', 'image radius',
                                 'lens1 info', 'lens2 info', '', '']
                                , sep_column, ""))

    if param['RMS_for_each_fld_wvl']:
        nb_wvl = len(pair_list[0].wvf_preprocessed[0]['wvl_and_data'])
        nb_fld = len(pair_list[0].wvf_preprocessed)
        file.write(data_list_to_str(['fld', 'wvl', 'RMS'] * nb_fld * nb_wvl, sep_column, ""))

    file.write(sep_row)

    pair_second_rms_sorted_index, sorted_rms_index_all = sort_pair_rms(pair_first_rms, pair_second_rms,
                                                                       index_min_rms_tolerance)

    # print pair infos line in the file
    for i_enum, i_pair in enumerate(sorted_rms_index_all):
        i_pair = int(i_pair)
        pair = pair_list[i_pair]

        if i_enum < len(pair_second_rms_sorted_index):  # if there is a second analysis RMS for this pair
            second_rms = pair_second_rms[pair_second_rms_sorted_index[i_enum]]
            # best RMS is the first element of sorted_rms_index_all, and beginning of sorted_rms_index_all
            # is just sorted_RMS_index_best

            if second_rms < 1 / 14:
                lambda14 = True
            else:
                lambda14 = False

        else:
            second_rms = None
            if pair_first_rms[i_pair] < 1 / 14:
                lambda14 = True
            else:
                lambda14 = False

        if hasattr(pair, 'foc_opti') and pair.foc_opti.success:
            bfl_opti = pair.bfl + pair.foc_opti.x[0]
        else:
            bfl_opti = ''  # None

        dist_first_last_surface_lens = opm_dist_idx1_idx2(pair.opm, pair.lens1surf[0], pair.lens2surf[1])
        dist_FFi = dist_first_last_surface_lens - pair.ffl + pair.bfl  # distance from object focal plane to image focal plane

        efl = pair.pm.opt_model['analysis_results']['parax_data'].fod.efl

        # if want anonymity of lenses path or not
        if param['output_csv']['hide_path_in_csv'] :
            tmp_path_lens1 = os.path.relpath(pair.lens1.filepath, param['lenses_param']['LensesFolderPath'])
            tmp_path_lens2 = os.path.relpath(pair.lens2.filepath, param['lenses_param']['LensesFolderPath'])
        else:
            tmp_path_lens1 = pair.lens1.filepath
            tmp_path_lens2 = pair.lens2.filepath

        if pair.lens1.filepath is None:
            tmp_filename_lens1 = None
        else:
            tmp_filename_lens1 = os.path.basename(pair.lens1.filepath)

        if pair.lens2.filepath is None:
            tmp_filename_lens2 = None
        else:
            tmp_filename_lens2 = os.path.basename(pair.lens2.filepath)

        if ('image_size' in param) and param['image_size']:
            img_radius_tmp = pair.img_radius['radius']
        else:
            img_radius_tmp = None

        file.write(
            data_list_to_str([tmp_filename_lens1, tmp_filename_lens2,
                              str(pair.config), pair.short_name, str(pair.lens1.id), str(pair.lens2.id),
                              pair.lens1.efl, pair.lens2.efl, pair_first_rms[i_pair],
                              second_rms, lambda14, pair.cutoffangle['value'], pair.ffl, pair.lenses_gap,
                              pair.bfl, bfl_opti, efl, dist_FFi, dist_first_last_surface_lens, tmp_path_lens1,
                              tmp_path_lens2,
                              pair.lens1.diameter_minmax[0], pair.lens1.diameter_minmax[1],
                              pair.lens2.diameter_minmax[0],
                              pair.lens2.diameter_minmax[1], img_radius_tmp, pair.lens1.info, pair.lens2.info,
                              i_pair, pair.rms_detail[0]['RMS']]
                             , sep_column, "", sep_dec))

        if param['RMS_for_each_fld_wvl']:
            nb_wvl = len(pair.wvf_preprocessed[0]['wvl_and_data'])
            for i_fld in range(len(pair.wvf_preprocessed)):  # fld
                for i_wvl in range(nb_wvl):  # wvl
                    tmp_fld = pair.rms_detail[i_fld * nb_wvl + i_wvl + 1]['used_fld']
                    tmp_wvl = pair.rms_detail[i_fld * nb_wvl + i_wvl + 1]['used_wvl']
                    tmp_rms = pair.rms_detail[i_fld * nb_wvl + i_wvl + 1]['RMS']

                    file.write(data_list_to_str([tmp_fld, tmp_wvl, tmp_rms], sep_column, "", sep_dec))

        file.write(sep_row)

    file.close()
    logger.info('results exported in : ' + file_path + '.csv')
    print('results exported in : ' + file_path + '.csv')


def save_results_csv_df(param, pair_dataframe):
    logger.info('saving results in csv ...')
    print('saving results in csv ...')

    # create the file
    file_path = os.path.join(param['save_folder_path'], param['analysis_name'])
    try:
        file = open(file_path + ".csv", "x")
    except:  # this should never happen due to precedent code, but if it is, we want to be sure we can create a new file
        i = 0
        while i < 100:
            i = i + 1
            try:
                file = open(file_path + "_" + str(i) + ".csv", "x")
                file_path = file_path + "_" + str(i)
                break
            except Exception as exception:
                raise ValueError('Impossible to create the csv result file. Error is : ' + str(exception))

    sep_row = param['output_csv']['output_csv_row_separator']
    sep_column = param['output_csv']['output_csv_column_separator']
    sep_dec = param['output_csv']['output_csv_decimal_separator']

    ######## HEADER ########
    # header
    file.write("Analysis of lens pairs for optical remote focusing." + sep_row)
    file.write(sep_row)
    file.write("Analysis name : " + param['analysis_name'] + sep_row)
    file.write("File save time : " + str(datetime.now()) + sep_row)
    file.write("All distances are in mm (except wavelengths : in nm), angles are in degrees,"
               " RMS wavefront error is in waves" + sep_row)
    file.write(sep_row)
    file.write("parameters :" + sep_row)
    for pa in param:
        if pa == 'save_folder_path' and param['output_csv']['hide_path_in_csv']:
            file.write(str(pa) + sep_column + 'hidden' + sep_row)
        elif pa == 'lenses_param' and param['output_csv']['hide_path_in_csv']:
            lenses_param_tmp = param[pa].copy()
            lenses_param_tmp['LensesFolderPath'] = 'hidden'
            file.write(str(pa) + sep_column + str(lenses_param_tmp) + sep_row)
        else:
            file.write(str(pa) + sep_column + str(param[pa]) + sep_row)

    file.write(sep_row)
    file.write("WARNING : if your lenses files don't take into account the reduction of lens diameter due to their"
               " mounting, the cut-off angle can be overestimated !" + sep_row)
    file.write("WARNING : Coating is not taken into account, please check if your lenses are designed for the "
               "right wavelengths !" + sep_row)
    file.write("WARNING : Always check if lenses were imported with correct materials, and if raytracing diagram "
               "and wavefront seems correct ! (especially when pair have negative ffl and/or bfl)" + sep_row)
    file.write(sep_row)

    ######## DATA ########
    # data
    file.write("pairs :" + sep_row)

    # column names
    file.write(data_list_to_str([pair_dataframe.index.name] + [col_name for col_name in pair_dataframe]
                                , sep_column, sep_row, sep_dec))

    # data
    for i_line in range(len(pair_dataframe.index)):
        file.write(data_list_to_str([pair_dataframe.index[i_line]] + pair_dataframe.iloc[i_line].to_list()
                                    , sep_column, sep_row, sep_dec))

    file.close()
    logger.info('results exported in : ' + file_path + '.csv')
    print('results exported in : ' + file_path + '.csv')


def sort_pair_rms(pair_first_rms, pair_second_rms, index_second_rms):
    # get pairs ordonated by RMS value
    if (not pair_second_rms is None) and pair_second_rms.size != 0:  # not empty
        # pair_second_rms are RMS of pairs of index index_min_rms_tolerance (in the same order)
        pair_second_rms_sorted_index = np.argsort(pair_second_rms)  # index to sort pair_second_rms
        sorted_rms_index_best = index_second_rms[pair_second_rms_sorted_index]  # id of pairs with the best RMS, sort by
                                                                                # the RMS in pair_second_rms
    else:
        index_min_rms_tolerance_sorted = []
        sorted_rms_index_best = np.array([], dtype=np.int64)
        pair_second_rms_sorted_index = np.array([], dtype=np.int64)

    # remove pairs that are in best index to avoid duplication
    pair_first_rms_without_second = np.delete(pair_first_rms, sorted_rms_index_best)
    convertion_table_rms_pairid = np.arange(0, len(pair_first_rms), 1)
    convertion_table_rms_pairid = np.delete(convertion_table_rms_pairid, sorted_rms_index_best)

    # sort by RMS
    sorted_rms_index_only_first = convertion_table_rms_pairid[np.argsort(pair_first_rms_without_second)]
    # id of pairs sorted by their RMS in first analysis

    # all pairs sorted
    if sorted_rms_index_best.size != 0:  # If empty. [] is float64 so issues with concatenate, even when force to int type with astype...
        sorted_rms_index_all = np.concatenate((sorted_rms_index_best, sorted_rms_index_only_first))
    else:
        sorted_rms_index_all = sorted_rms_index_only_first

    return pair_second_rms_sorted_index, sorted_rms_index_all


def export_detail_pair_seqmod_txt(pair, param=None, overwrite_file=False, filename=None, dir_path=None):
    # export details about surfaces and gaps of the pair, write them in a file ".txt"
    # overwrite_file is to say if a file has the same path, should the function overwrite it ?
    # filename : if None, filename will be param['analysis_name']_pair'pair id'_SeqMod
    # dir_path : path of the folder, if None, dir_path = param['save_folder_path']
    # param can be None only if filename and dir_path aren't None

    logger.info('start to save pair surfaces and gaps information : ' + pair.name())

    if (param is None) and (filename is None or dir_path is None):
        logger.error('Error in export_detail_pair_SeqMod_txt for pair' + pair.name() +
                     "\nparam can be None only if filename AND dir_path aren't None")
        return

    # create the file
    if filename is None:
        filename = param['analysis_name'] + "_pair" + str(pair.id) + "_SeqMod"
    if dir_path is None:
        dir_path = param['save_folder_path']

    file_path = os.path.join(dir_path, filename)

    if not overwrite_file:
        try:
            file = open(file_path + ".txt", "x")
        except:  # we want to be sure we can create a new file
            i = 0
            while i < 100:
                try:
                    file = open(file_path + "_" + str(i) + ".txt", "x")
                    file_path = file_path + "_" + str(i)
                    break
                except:
                    i = i + 1

            if i >= 100:
                logger.error('Error in export_detail_pair_SeqMod_txt for pair' + pair.name() +
                             '\nfilename is already taken by another file, and impossible to find a new filename' +
                             '\nconsider running again the function but with a new, free filename')
                return
    else:
        file = open(file_path + ".txt", "w")

    # write in the file
    file.write("Analysis of lens pairs for optical remote focusing.\n\n" +
               "Analysis name : " + param['analysis_name'] + "\n" +
               "File save time : " + str(datetime.now()) + "\n" +
               "All distances are in mm (except wavelengths : in nm)\n\n" +
               "Details about sequential model of pair :\n" + pair.name() + '\n' +
               "Pair short name : " + pair.short_name + "\n\n")

    file.write(detail_seq_model_str(pair))

    logger.info('Save file ' + file_path + '.txt')


def detail_seq_model_str(pair):
    more_info = ''
    more_info = more_info + "List sequential model :\n"

    # capture the output of listing sequential model (which is output with print, not in str)
    # https://stackoverflow.com/questions/22822267/how-to-capture-print-output-of-another-module
    # https://docs.python.org/3/library/contextlib.html
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        pair.sm.list_model()
    list_model_str = f.getvalue()
    more_info = more_info + list_model_str + "\n\n"

    more_info = more_info + "wavelengths information :\n" + pair.osp.spectral_region.listobj_str() + "\n"

    more_info = more_info + "First order properties :\n"

    f2 = io.StringIO()
    with contextlib.redirect_stdout(f2):
        pair.pm.first_order_data()
    first_order_data_str = f2.getvalue()
    more_info = more_info + first_order_data_str + "\n\n"

    more_info = more_info + ("More information about surfaces and gaps (coefficients if some surfaces are asphere, "
                             "refractive index used, and more) :\n")

    for i in range(len(pair.sm.ifcs)):
        if i == pair.lens1surf[0]:
            more_info = more_info + "=========== LENS 1 START ===========\n"
        elif i == pair.lens2surf[0]:
            more_info = more_info + "=========== LENS 2 START ===========\n"

        more_info = (more_info + "INTERFACE " + str(i) + " :\n" +
                     pair.sm.ifcs[i].listobj_str())

        if i == pair.lens1surf[1]:
            more_info = more_info + "=========== LENS 1 END ===========\n\n"
        elif i == pair.lens2surf[1]:
            more_info = more_info + "=========== LENS 2 END ===========\n\n"
        else:
            more_info = more_info + '\n'

        if i < len(pair.sm.ifcs) - 1:
            more_info = (more_info + "GAP " + str(i) + " :" +
                         pair.sm.gaps[i].listobj_str())
            for wvl_weight in pair.initial_wvl_and_weight:
                wvl, weight = wvl_weight
                more_info = more_info + "RI of medium for wvl " + str(wvl) + " nm is : " + str(
                    pair.sm.gaps[i].medium.calc_rindex(wvl)) + "\n"
            more_info = more_info + "\n"
    return more_info




import logging
import os  # to determine files path

import pandas as pd
import numpy as np
from .util import Singleton
from opticalglass import glass
from opticalglass.spectral_lines import get_wavelength


logger = logging.getLogger('PyRFTL_opticalglass_schott')


class SchottCatalog(metaclass=Singleton):
    def __init__(self, fname=None):
        if fname is None:
            file_dir = os.path.dirname(os.path.realpath(__file__))
            fname = os.path.join(file_dir, r'data\schott-optical-glasses-preferred-and-inquiry-en.AGF')

        # Open AGF schott file, extract data about dispersion formula

        # create a dataframe to stock data. Each line = one material
        self.df = pd.DataFrame(columns=['formula_number',
                                        'c0','c1','c2','c3','c4','c5','c6','c7','c8','c9',
                                        'wvl_min_nm', 'wvl_max_nm'])
        # glass name => index

        # open the agf file and read it line by line
        # fill the dataframe self.df with informations about materials
        with open(fname) as agf_file :
            cur_mat = None
            suppress_mat = []  # materials to delete because error when import them
            for line in agf_file :
                line_elements = line.split(' ')

                match line_elements[0] :
                    case 'NM':
                        # self.df add a line (new material)
                        cur_mat = line_elements[1]
                        self.df.loc[cur_mat, 'formula_number'] = line_elements[2]

                    case 'CD':
                        # self.df add diffusion coefficients to current line = last material added
                        nb_elts = len(line_elements)
                        number_last_coeff = str(nb_elts - 2 )
                        try :
                            self.df.loc[cur_mat, 'c0' : 'c'+number_last_coeff] = [float(coeff) for coeff in
                                                                                  line_elements[1:nb_elts]]

                        except Exception as exception :
                            # wrong import of coefficients
                            self.df.loc[cur_mat, 'c0': 'c' + number_last_coeff] = [np.nan]*(nb_elts-1)
                            # logger.error('Error when import materials from file : ' + fname +
                            #              '\nError with material ' + cur_mat + '. When try to import its coefficients : '
                            #              + str(exception))
                            if cur_mat not in suppress_mat :
                                suppress_mat.append(cur_mat)

                    case 'LD':
                        try :
                            self.df.loc[cur_mat, 'wvl_min_nm'] = float(line_elements[0]) * 1e3
                            self.df.loc[cur_mat, 'wvl_max_nm'] = float(line_elements[1]) * 1e3
                        except:
                            pass

                    case _ :
                        pass

            # delete materials with bad import
            for mat_to_delete in suppress_mat :
                self.df = self.df.drop(mat_to_delete)

        # some values
        self.name = 'Schott'

        # build an alphabetical list of decoded glass names
        gnames = self.df.index.array
        glass_list = [(glass.decode_glass_name(gn), gn, self.name)
                      for gn in gnames]
        glass_list = sorted(glass_list, key=lambda glass: glass[0][0])
        # build a lookup dict of the glass defs keyed to decoded glass names
        glass_lookup = {gn_decode: (gn, gc)
                        for gn_decode, gn, gc in glass_list}
        # attach these 'static' lists to class variables
        self.__class__.glass_list = glass_list
        self.__class__.glass_lookup = glass_lookup

    def get_glass_names(self):
        """ returns a list of glass names """
        return self.df.index.array

    def get_column_names(self):
        """ returns a list of column headers """
        return self.df.columns.levels[0]

    def glass_index(self, gname):
        """ returns the glass index (row) for glass name `gname`
        Args:
            gname (str): glass name
        Returns:
            int: the 0-based index (row) of the requested glass
        """
        return self.df.index.get_loc(gname)

    def glass_coefs_zmx(self, gname):
        """ returns an array of glass coefficients for the glass at *gname*
        coefficients order is the one of the zmx file (line CD) """
        glas = self.df.loc[gname]
        coefs = glas['c0':'c9'].to_numpy(dtype=float)
        return coefs

    def glass_coefs(self, gname):
        """ returns an array of glass coefficients for the glass at *gname*
        coefficient order is the same than used by rayoptics schott function based on xls file, for compatibility"""
        c = self.glass_coefs_zmx(gname)
        coefs = [c[0], c[2], c[4], c[1], c[3], c[5], c[6], c[7], c[8], c[9]]
        return coefs

    def create_glass(self, gname: str, gcat: str) -> 'SchottGlass':
        """ Create an instance of the glass `gname`. """
        return SchottGlass(gname)

    def catalog_name(self):
        return self.name


class SchottGlass():
    catalog = None

    def initialize_catalog(self):
        if SchottGlass.catalog is None:
            SchottGlass.catalog = SchottCatalog()

    def __init__(self, gname):
        self.initialize_catalog()
        self.gname = gname
        self.coefs = self.catalog.glass_coefs(gname)  # coeff order of RayOptics SchottGlass class using xls file
        self.coefs_zmx = self.catalog.glass_coefs_zmx(gname)  # coeff order of zemax file line CD

    def calc_rindex(self, wv_nm):
        # return the refractive index of the medium for the wavelength wv_nm (should be in nm)

        if not hasattr(self, 'coefs_zmx'):
            # compatibility with medium created with unmodified version of opticalglass
            # if the medium is a schott glass of an unmodified version, change it to the modified SchottGlass
            # can be useful in rayoptics when open a roa file
            self = SchottGlass(self.gname)

        wv = 0.001 * wv_nm  # convert in um
        wv2 = wv * wv
        c = self.coefs_zmx
        n2 = 1. + (c[0] * wv2 / (wv2 - c[1])) + (c[2] * wv2 / (wv2 - c[3])) + (c[4] * wv2 / (wv2 - c[5]))
        return np.sqrt(n2)

    def rindex(self, wvl) -> float:
        """ returns the interpolated refractive index at wvl

        Args:
            wvl: either the wavelength in nm or a string with a spectral line
                 identifier. for the refractive index query

        Returns:
            float: the refractive index at wv_nm
        """
        return self.calc_rindex(get_wavelength(wvl))

    def name(self):
        """ returns the glass name, :attr:`gname` """
        return self.gname

    def __repr__(self):
        return "{!s}('{}')".format(type(self).__name__, self.gname)

    def catalog_name(self):
        """ returns the glass name, :attr:`gname` """
        return self.catalog.catalog_name()


sc = SchottCatalog()



from pyrftl.func.Lens import Lens
from pyrftl.func.pairs import Pair
from pyrftl.func.wavefront_pair_analysis import wvf_fast_one_pair


def analyse_one_pair (param_one):
    lens1 = Lens(param_one['LensFile1'], check_air_gap=param_one['check_air_gap'], delete_air_gap=param_one['delete_air_gap'])
    lens2 = Lens(param_one['LensFile2'], check_air_gap=param_one['check_air_gap'], delete_air_gap=param_one['delete_air_gap'])
    pair = Pair(lens1, lens2, param_one['expected_focal_length'])
    pair.create_optical_model(config=param_one['config'][0], wvl_and_weight=param_one['wvl_and_weight'],
                              pupil=param_one['pupil'], fld=[0], dist_pupil_pair=param_one['dist_pupil_pair'])
    if param_one['cut_off_angle_test_end'] is None:
        if param_one['cut_off_angle_min'] < 35 :
            param_one['cut_off_angle_test_end'] = 45
        else :
            param_one['cut_off_angle_test_end'] = param_one['cut_off_angle_min'] * 2
    if param_one['cut_off_angle_test_end'] < param_one['cut_off_angle_min'] :
        param_one['cut_off_angle_test_end'] = param_one['cut_off_angle_min'] + 2*param_one['cut_off_angle_precision']

    pair.determine_cutoffangle_dichotomy(angle_end=param_one['cut_off_angle_test_end'],
                                         angle_precision=param_one['cut_off_angle_precision'])
    cut_off = pair.cutoffangle['value']

    wvf_fast_one_pair(pair, param_one, param_one['grid'])
    rms = pair.rms_detail

    return pair, cut_off, rms




import pyrftl.func.lens_selection as fls


if __name__ == '__main__':

    ######### USER PARAMETERS #########
    LensesFolderPath = r"PATH\lenses\Lenses_achromats"

    ######### ADVANCED USER PARAMETERS #########
    check_air_gap = True
    delete_air_gap = False

    lens_diameter_min = None
    lens_diameter_max = None
    lens_efl_min = None
    lens_efl_max = None

    ######### USER PARAMETERS #########
    # expected_focal_length = 162.4
    expectedFocalLength = fls.f_tl2_from_f_tl1(60, 180, 180, 50, 180, 1.33, 1)

    distMinBetweenLenses = None  # mm
    distMinBeforeLenses = None
    distMinAfterLenses = None

    distMaxBetweenLenses = None
    distMaxBeforeLenses = None
    distMaxAfterLenses = None

    wvl = [486, 588, 656]  # in nm, wavelengths : can be one : wvl = [540] or several wvl = [486, 588, 656]
    # pupil = 6.84  # diameter, in mm
    pupil = fls.pupil_diameter(0.95, 50, 180, 1.33, 1.2)  # 6.5
    # cut_off_angle_min = 4.2  # in °, maximum field angle that should be able to pass inside the tube length
    cut_off_angle_min = fls.max_fld_angle(26.5, 50, 180, 1.33, 26.5, 60)  # 4.21

    config = ["A","B","C","D"]
    # config can be only one kind : config = "A", or it is possible to use several configs : config = ["A", "B", "D"]
    # config define what are lenses orientation in the pair. With ( the more curved side of the lens, | the less curved
    # side, and the objective in remote refocus is at the infinite side, we have :
    # "A" : infinity |) |) image    "B" : infinity |) (| image
    # "C" : infinity (| (| image    "D" : infinity (| |) image
    # None : lenses orientations are the ones of the file from where they are obtained

    save_Folder_Path = r'SAVEPATH'  # path of the folder where results should be save
    analysis_name = None  # should be None or a str. If None, will be : PyRFTL_year_month_day_hourhminute

    dist_pupil_pair = None  # in mm, distance between the pupil and the first lens.
    # if None, distance will be pair front focal length
    # for remote focusing and Oblique Plane Microscopy, put None

    ######### ADVANCED USER PARAMETERS #########
    thin_lens_first_analysis = True  # if True, will do a first selection of pairs with a thin lens analysis,
                                     # for distances criterias.
                                     # In any case, will definitively select pairs with ray tracing.
    accepted_error_in_dist_thin_lens_percent = 10  # in percent, between 0 and 100.

    number_best_pair_high_comparison = 10
    # None if keep all with RMS > tolerance * lambda/14, if an integer, keep the given number of pairs that have
    # minimum RMS, and have RMS > tolerance * lambda/14
    # The maximum number of best pairs that are selected for an RMS comparison with more rays
    tolerance_wavefront_RMS_high = 10  # in percent, between 0 and 100

    nb_subprocesses = None  # The number of subprocesses to use in parallel for multiprocessing : int or None
    # if None, it will be equal to the number of cores of the CPU
    force_affinity_core = True  # in some case, the multiprocessing does not use all cpu cores it should.
    # If it is the case on your computer, you can put force_affinity_core to True. It will try to force the use of
    # a different cpu core for each subprocess. Some stuff took time to be loaded by cores, so even with this parameter,
    # your CPU will not be at 100% usage all the time

    output_csv_column_separator = ';'
    output_csv_row_separator = '\n'
    output_csv_decimal_separator = '.'
    hide_path_in_csv = False
    # bool to say if should hide or not the main path of lenses in output csv (should just show the path after
    # folder_lenses_path if True)
    # WARNING !!! even with this parameter, complete path will be output in log file. hide only in csv output file

    save_roa = 'best'  # 'no', 'best', 'all'
    # save rayoptics optical model : 'no' : don't save, 'best' : save only pairs with best rms (rms<1/14, and
    # in the top number_best_pair_high_comparison number of pairs)
    # 'all' : save it for all pairs

    save_seq_detail = 'best'  # 'no', 'best', 'all'
    # save details about sequential model in a txt file : 'no' : don't save, 'best' : save only pairs with best rms (rms
    # <1/14, and in the top number_best_pair_high_comparison number of pairs)
    # 'all' : save it for all pairs
    # it is recommended to manually check if pairs you want to use were well generated, so to use 'best' or 'all'

    ######### REALLY ADVANCED USER PARAMETERS DO NOT MODIFY IF NO GOOD REASON #########
    best_focus_fast = True
    # If True, put the image at RMS lower wavefront error plane for field 0°, if False, put the image at bfl
    nbRays_first_comparison = 16
    nbRays_high_comparison = 64

    cut_off_angle_test_end = None
    # cut-off angle higher than this value will not be calculated and will be considered equal to this value
    # should be higher than fldAngleMin. If cut_off_angle_test_end < fldAngleMin, it will be set :
    # cut_off_angle_test_end = fldAngleMin + 2*cut_off_angle_precision
    # by default, with "cut_off_angle_test_end = None", it is set : cut_off_angle_test_end = 45° if fldAngleMin < 35°,
    # fldAngleMin*2 else

    cut_off_angle_precision = 0.1  # in °

    polychromatic = 'rmse_rms'
    # in the case of multi-wavelengths, formula to use as criteria
    # 'rmse_rms' is the root mean square of the root mean square error of wavelength wavefront aberrations
    # 'worst' is the worst root mean square error of wavefront aberrations

    focus_polychromatic = 'ref_wvl_i'
    # in multi-wavelength case, how to set the image position for the 0° field, before getting the RMSE at max field angle
    # 'ref_wvl_i' align with the reference wavelength
    # 'rmse_rms' align with the RMS of RMSE of wavelengths wavefront errors

    save_ram = "max"
    # "no" "medium" "max"
    # to use less computer memory while running the software. "no" is not recommended if you compute on a lot of lenses.

    wvl_and_weight = None
    # wvl_and_weight : if None, will use wvl, and give a weight of 1 for all wavelengths.
    # weight is currently not implemented, so it is currently useless to put something else than None for wvl_and_weight.
    # to define wvl and weight, it should be on the shape : [(wvl1,weight1),(wvl2,weight2),etc]
    # for example : [(510,1), (560,1), (630,2)]




    ###### NO MORE USER PARAMETERS BELOW THIS LINE DO NOT MODIFY #####
    from pyrftl.select_pair_successive import select_pair
    from pyrftl.func.save import create_save_folder

    if wvl_and_weight is None:
        if not isinstance(wvl, list):
            wvl = [wvl]

        wvl_and_weight = []

        for wvl_i in wvl:
            wvl_and_weight.append((wvl_i, 1))

    if nb_subprocesses is None :
        import multiprocess as mp
        nb_subprocesses = mp.cpu_count()

    complete_save_Folder_Path, analysis_name, start_time = create_save_folder(save_Folder_Path, analysis_name)

    dist_criteria = {'O1O2_min': distMinBetweenLenses, 'O1O2_max': distMaxBetweenLenses, 'ffl_min': distMinBeforeLenses,
                     'ffl_max': distMaxBeforeLenses, 'bfl_min': distMinAfterLenses, 'bfl_max': distMaxAfterLenses,
                     'tolerance_prcent': accepted_error_in_dist_thin_lens_percent}

    lenses_param = {'LensesFolderPath':LensesFolderPath, 'check_air_gap':check_air_gap, 'delete_air_gap':delete_air_gap,
                    'lens_diameter_min':lens_diameter_min, 'lens_diameter_max':lens_diameter_max,
                    'lens_efl_min':lens_efl_min,'lens_efl_max':lens_efl_max}

    output_csv = {'output_csv_column_separator': output_csv_column_separator,
                  'output_csv_row_separator': output_csv_row_separator, 'hide_path_in_csv': hide_path_in_csv,
                  'output_csv_decimal_separator': output_csv_decimal_separator}

    param = {'expected_focal_length': expectedFocalLength, 'dist_criteria': dist_criteria, 'lenses_param' : lenses_param,
             'thin_lens_first_analysis': thin_lens_first_analysis,
             'save_ram': save_ram, 'config': config, 'wvl_and_weight': wvl_and_weight, 'pupil': pupil,
             'dist_pupil_pair' : dist_pupil_pair, 'cut_off_angle_min': cut_off_angle_min,
             'cut_off_angle_test_end': cut_off_angle_test_end, 'cut_off_angle_precision': cut_off_angle_precision,
             'best_focus_fast': best_focus_fast, 'nbRays_first_comparison': nbRays_first_comparison,
             'nbRays_high_comparison': nbRays_high_comparison,
             'number_best_pair_high_comparison': number_best_pair_high_comparison,
             'tolerance_wavefront_RMS_high_percent': tolerance_wavefront_RMS_high,
             'save_folder_path': complete_save_Folder_Path, 'analysis_name': analysis_name,
             'nb_subprocesses' : nb_subprocesses,
             'polychromatic' : polychromatic, 'focus_polychromatic' : focus_polychromatic,
             'force_affinity_core' : force_affinity_core, 'output_csv': output_csv,
             'save_roa': save_roa, 'save_seq_detail': save_seq_detail,
             'start_time':start_time}

    pair_list = select_pair(param)




# import from general libraries
from glob import glob  # to get all files path from a folder path
import os  # to get filename form path
import logging

import numpy as np

from configparser import ConfigParser  # import txt file *.ini to import lens folder metadata

# import from PyRFTL
from pyrftl.func.Lens import Lens
logger = logging.getLogger('PyRFTL')


def open_lenses (lenses_param, recursive=True) :
    # this function return a lenses list from lenses opened in lenses_param['LensesFolderPath']

    list_zmx_files = glob(lenses_param['LensesFolderPath'] + "/**/*.zmx", recursive=recursive)
    list_codev_files = glob(lenses_param['LensesFolderPath'] + "/**/*.seq", recursive=recursive)
    list_ray_optics_files = glob(lenses_param['LensesFolderPath'] + "/**/*.roa", recursive=recursive)

    lenses_filepath_list = [*list_zmx_files, *list_codev_files, *list_ray_optics_files]  # concatenate
    lenses_list = []

    logger.info("Start to import lenses...\n")
    print("Try to open lens 1/" + str(len(lenses_filepath_list)), end='')
    j = 0

    # try to open each lens in the folder, and to get main properties
    for i, lensFilePath in enumerate(lenses_filepath_list):
        print("\rTry to open lens " + str(i + 1) + "/" + str(len(lenses_filepath_list)), end='')
        try:
            try:  # get lens folder metadata
                metadata_path = os.path.join(os.path.dirname(lensFilePath), 'lens_dir_metadata.txt')

                if os.path.exists(metadata_path):
                    # if there is a metadata file in the lens folder, extract its data
                    metadata = ConfigParser()
                    metadata.read(metadata_path)
                    info = metadata._sections.copy()  # extract the dictionary
                else:
                    info = None
            except:
                info = None

            # open lenses
            lens = Lens(lensFilePath, lenses_param['check_air_gap'], lenses_param['delete_air_gap'], None, info)

            # check if lens mets users parameters
            add_lens = True

            # check diameter
            if not lenses_param['lens_diameter_min'] is None:
                if lens.diameter_minmax[0] < lenses_param['lens_diameter_min'] :  # min diameter too small
                    add_lens = False
                    why_not_add_lens = ('This lens is rejected because its diameter is too small, change lens_diameter_min'
                                        + ' if you want to keep it \nThis lens has a min diameter of :  ' +
                                        str(lens.diameter_minmax[0]) + ' mm')

            if (not lenses_param['lens_diameter_max'] is None) and add_lens:
                if lens.diameter_minmax[1] > lenses_param['lens_diameter_max']:  # max diameter too huge
                    add_lens = False
                    why_not_add_lens = ('This lens is rejected because its diameter is too large, change lens_diameter_max'
                                        + ' if you want to keep it \nThis lens has a max diameter of :  '
                                        + str(lens.diameter_minmax[1]) + ' mm')

            # check effective focal length
            if (not lenses_param['lens_efl_min'] is None) and add_lens:
                if lens.efl < lenses_param['lens_efl_min']:  # too small efl
                    add_lens = False
                    why_not_add_lens = ('This lens is rejected because its effective focal length is too short, change'
                                        + ' lens_efl_min if you want to keep it \nThis lens efl is :  '
                                        + str(lens.efl) + ' mm')

            if (not lenses_param['lens_efl_max'] is None) and add_lens:
                if lens.efl > lenses_param['lens_efl_max']:  # too huge efl
                    add_lens = False
                    why_not_add_lens = ('This lens is rejected because its effective focal length is too long, change '
                                        + 'lens_efl_max if you want to keep it \nThis lens efl is :  '
                                        + str(lens.efl) + ' mm')

            # keep the lens if it is ok
            if add_lens:
                lens.id = j
                j = j + 1
                lenses_list.append(lens)
                logger.info("Lens selected for generation of pairs : " + os.path.basename(lens.filepath) + "\n")
            else:
                logger.info("Lens is rejected :\n" + lensFilePath + "\n" + why_not_add_lens + "\n")

        except Exception as exception:
            logger.error("Error while trying to open file : " + lensFilePath + "\n" +
                         "Error is : " + str(exception))
            logger.info("Software is still running, opening next lens\n")
            raise

    print()
    return lenses_list


def max_fld_angle(fn_O2, M_O2, ftl_O2m, Mtot=None, fn_O1=None, M_O1=None):
    # return the maximum field angle (in degree) after infinity corrected objective.
    # can work with only one objective (Mtot=None, fn_O1=None, M_O1=None) or with a remote refocusing system
    # (see Botcherby et al. paper doi:10.1364/OL.32.002007)
    # (see also Hong et al. https://doi.org/10.1364/OE.450320)
    # In the second case, we give the maximum field angle after O2, with TL2 is the tube lens we want to design
    #
    # fn_O2 field number of O2
    # M_O2 initial magnification of O2
    # ftl_O2m tube lens focal length for which O2 is designed
    #
    # in case of optical remote focusing
    # Mtot is the magnification from the sample (before O1) to the remote sample (after O2), usually n1/n2
    # fn_O1 field number of objective O1
    # M_O1 initial magnification of O1
    # to get max field angle for only one objective, and not in remote refocusing configuration, put
    # Mtot=None, fn_O1=None, M_O1=None

    sample_diam_2 = fn_O2/M_O2

    if not (Mtot is None or fn_O1 is None or M_O1 is None) :
        sample_diam_1 = Mtot * fn_O1/M_O1
        sample_diam = min(sample_diam_1, sample_diam_2)
    else:
        sample_diam = sample_diam_2

    fld_angle_max = np.arctan(sample_diam * M_O2 / (2 * ftl_O2m))
    return fld_angle_max * 180/np.pi  # degree


def pupil_diameter (NA_O2, M_O2, ftl_O2m, Mtot=None, NA_O1=None):
    # return the pupil diameter of the objective infinity corrected objective O2
    # can work with only one objective (Mtot=None, NA_O1=None, M_O1=None) or with a remote refocusing system
    # (see Botcherby et al. paper doi:10.1364/OL.32.002007)
    # (see also Hong et al. https://doi.org/10.1364/OE.450320)
    # In the second case, we give the pupil diameter of O2, with TL2 is the tube lens we want to design
    #
    # NA_O2 numerical aperture of O2
    # M_O2 initial magnification of O2
    # ftl_O2m tube lens focal length for which O2 is designed
    #
    # in case of optical remote focusing
    # Mtot is the magnification from the sample (before O1) to the remote sample (after O2), usually n1/n2
    # NA_O1 numerical aperture of objective O1
    # M_O1 initial magnification of O1
    # to get max field angle for only one objective, and not in remote refocusing configuration, put
    # Mtot=None, NA_O1=None, M_O1=None

    p2 = 2 * NA_O2 * ftl_O2m/M_O2

    if not (Mtot is None or NA_O1 is None):
        p1_M = 2 * NA_O1 * ftl_O2m/(Mtot * M_O2)
        p = min(p2, p1_M)
    else:
        p = p2

    return p


def f_tl2_from_f_tl1 (MO1_m, ftl1_m, ftl1_u, MO2_m, ftl2_m, n1, n2):
    # in case of optical remote focusing system as described in Botcherby paper doi:10.1364/OL.32.002007,
    # in the case TL1 is fixed, and we are looking for the required focal length for TL2
    # this function return the required focal length for TL2
    #
    # MO1_m, MO2_m manufacturer magnification of objectives O1 and O2
    # ftl1_m, ftl2_m manufacturer focal length of tubes lengths required for O1 and O2
    # ftl1_u used focal length for TL1
    # n1, n2 refractive index before O1, after O2 (required magnification for the system will be Mtot = n1/n2)

    ftl2_u = ftl1_u * ftl2_m * MO1_m * n2 / (n1 * MO2_m * ftl1_m)
    return ftl2_u





# import from general libraries
from rayoptics.environment import *  # to import RayOptics. Import also matplotlib.pyplot as plt

from scipy import optimize  # optimization functions
import numpy as np

import os  # to get filename form path

from copy import deepcopy  # to copy an object
import gc  # to free memory

import logging

import contextlib, io  # to be able to capture stdout entry, so what is print(...) by a module

from decimal import Decimal, ROUND_DOWN  # to round down cut-off angle to desire precision, avoid 30 nonsense digits...

# import matplotlib.pyplot as plt

# import from PyRFTL
from pyrftl.rayopt.copy_opm import copy_opm_internal, set_clear_apertures
from pyrftl.func.various_stuff import obtain_distance_to_cross_a_zero, lens_distance_to_efl_zero, safe_lens_distance_to_efl
from pyrftl.func.Lens import Lens
from pyrftl.rayopt import analyses_test as ta

logger = logging.getLogger('PyRFTL')


class ThinlensPairAnalysis:
    # this class is intended to give the distance between the two lenses of the pair in the case these lenses are thin lens

    def __init__(self, lens1_fl, lens2_fl, expected_fl):
        # lens1_fl and lens2_fl are respectively the focal length of lens 1 and 2
        # expected_fl is the combine focal length of both lenses
        self.lens1_fl = lens1_fl
        self.lens2_fl = lens2_fl
        self.expectedFL = expected_fl

        if lens2_fl is not None:
            [d, FO1, O2Fi, FFi] = self.thin_lens_pair_dist(lens1_fl, lens2_fl, expected_fl)
        else:
            [d, FO1, O2Fi, FFi] = [lens1_fl, lens1_fl, lens1_fl, 2 * lens1_fl]
        self.O1O2 = d
        self.FO1 = FO1
        self.O2Fi = O2Fi
        self.FFi = FFi

    @staticmethod
    def thin_lens_pair_dist(fi1, fi2, fitot):
        # use two lens L1 and L2 as an equivalent lens Ltot
        # this script return the distance from O1 to Ftot and O2 to F'tot, also distance
        # O1O2 between the two lenses
        #
        # input : focal lens first lens fi1
        # focal lens second lens fi2
        # desired focal lens of combination : fitot
        #
        # output : Separation between lenses O1O2 : d
        # distance focal to first lens FO1
        # distance last lens to focal O2F' (O2Fi)
        # distance between two focal FF'

        d = fi1 + fi2 - fi1 * fi2 / fitot
        O1F = fi1 * ((fi2 - d) / (d - fi1 - fi2))
        O2Fi = fi2 * ((d - fi1) / (d - fi1 - fi2))
        FO1 = -O1F
        FFi = d - O1F + O2Fi

        return [d, FO1, O2Fi, FFi]

    def list_attributes_str(self):
        list_attributes = ("Lens 1 FL = " + str(self.lens1_fl) + " mm\nLens 2 FL = " + str(self.lens2_fl) +
                           " mm\nexpected FL = " + str(self.expectedFL) + " mm\nO1O2 = " + str(self.O1O2) +
                           " mm\nFO1 = " + str(self.FO1) + " mm\nO2Fi = " + str(self.O2Fi) + " mm\nFFi = " + str(
                    self.FFi) + " mm")
        return list_attributes


class Pair:
    def __init__(self, lens1, lens2=None, expected_fl=None, pair_id=None, lens_info=None, copy_lens=True):
        # lens1 & lens2 are of class Lens
        # if lens2 is None and expected_fl is None, it is just one lens system

        if (lens2 is None) ^ (expected_fl is None):  # xor
            try:
                if lens2.filepath is None and type(lens2) is Lens :
                    pass
                else:
                    raise ValueError('lens2 and expected_fl should either both have a value, or both be None.'
                                     'You should not have only one defined as None.')
            except:
                raise ValueError('lens2 and expected_fl should either both have a value, or both be None.'
                                 'You should not have only one defined as None.')

        # lens1 is just a link to lens1 variable. If risk to modify it, it is better to get a copy that will not change,
        # use copy_lens = True for it. This will use more memory.
        if copy_lens:
            self.lens1 = deepcopy(lens1)
            self.lens2 = deepcopy(lens2)
        else:
            self.lens1 = lens1
            self.lens2 = lens2

        if self.lens2 is None:
            self.lens2 = Lens(None)

        self.expected_fl = expected_fl  # expected focal length for the pair

        self.thin_lens_analysis = ThinlensPairAnalysis(self.lens1.efl, self.lens2.efl, expected_fl)

        self.id = pair_id
        self.short_name = str(lens1.id) + '_' + str(self.lens2.id)
        self.lens_info = lens_info

        self.avoid_surface_change_list = []

    ##################################################      OPTICAL MODEL      ##################################################

    def create_optical_model(self, config="None", wvl_and_weight=None, ref_wvl_i=None, pupil=6.86, fld=None,
                             dist_pupil_pair=None):
        # config="None" or None if put lenses without changing their orientation from Lens class
        # config = "A", "B", "C", "D" see https://doi.org/10.1364/OE.450320
        # dist_pupil_pair is the distance between the entrance aperture and the first lens of the pair.
        # if dist_pupil_pair = None, it would be the ffl, else distance in mm
        # ref_wvl_i is the index of the reference wavelength in wvl_and_weight list. Default is 0.

        # wvl_and_weight and fld should be list. if None, default values will be applied
        if wvl_and_weight is None :
            wvl_and_weight = [('F', 1.0), ('d', 1.0), ('C', 1.0)]  # 486.1327 nm, 587.5618 nm, 656.2725 nm
            ref_wvl_i = 1

        if fld is None :
            fld = [0]

        if ref_wvl_i is None:
            ref_wvl_i = 0

        if not isinstance(fld, list):
            fld = [fld]

        self.config = config
        self.short_name = self.short_name + '_' + str(config)

        # create the optical model
        self.opm = OpticalModel()
        self.sm = self.opm['seq_model']
        self.osp = self.opm['optical_spec']
        self.pm = self.opm['parax_model']
        self.em = self.opm['ele_model']
        self.pt = self.opm['part_tree']
        self.ar = self.opm['analysis_results']

        self.opm.radius_mode = True
        self.sm.do_apertures = False
        self.sm.gaps[0].thi = 1e10  # object distance = infinity

        self.osp['wvls'] = WvlSpec(wvl_and_weight, ref_wl=ref_wvl_i)
        self.osp['pupil'] = PupilSpec(self.osp, key=['object', 'epd'], value=pupil)
        self.osp['fov'] = FieldSpec(self.osp, key=['object', 'angle'], flds=fld, is_relative=False)

        self.initial_wvl_and_weight = wvl_and_weight
        self.initial_ref_wvl_index = ref_wvl_i
        self.initial_fld_list = fld

        # add diaphragm pupil
        self.sm.add_surface([0, 0], sd=0.5*pupil)
        self.sm.set_stop()
        self.avoid_surface_change_list = [self.sm.cur_surface]
        self.update_model()

        # add gap before lenses
        self.sm.add_surface([0, 0])
        self.update_model()

        # add lenses
        self.lens1surf = [self.sm.cur_surface + 1]  # ext surface number of lens 1
        self.opm.add_part(copy_opm_internal, self.lens1.opm, t=self.thin_lens_analysis.O1O2)
        self.lens1surf.append(self.sm.cur_surface)

        # avoid to modify the lens 1 when update the optical model.
        self.avoid_surface_change_list = (self.avoid_surface_change_list +
                                          [i for i in range(self.lens1surf[0], self.lens1surf[1]+1)])  # concatenate

        if self.lens2.efl is not None :
            self.lens2surf = [self.sm.cur_surface + 1]
            self.opm.add_part(copy_opm_internal, self.lens2.opm, t=self.thin_lens_analysis.O2Fi)
            self.lens2surf.append(self.sm.cur_surface)

            # avoid to modify the lens 2 when update the optical model
            self.avoid_surface_change_list = (self.avoid_surface_change_list +
                                              [i for i in range(self.lens2surf[0], self.lens2surf[1] + 1)])
        else:
            self.lens2surf = [self.sm.cur_surface, self.sm.cur_surface]

        self.update_model()

        self.elmn = [node.id for node in self.pt.nodes_with_tag(tag='#element')]

        # change lenses direction to match with the desired configuration
        match config:
            case "A" | "a":
                # put lenses with more curved side in the image side
                # Object (infinite) lens1 |) lens 2 |) image
                if self.lens1.curv_state == "first_curv":
                    self.opm.flip(self.lens1surf[0], self.lens1surf[1])
                if self.lens2.curv_state == "first_curv":
                    self.opm.flip(self.lens2surf[0], self.lens2surf[1])

            case "B" | "b":
                # infinity |) (| image
                if self.lens1.curv_state == "first_curv":
                    self.opm.flip(self.lens1surf[0], self.lens1surf[1])
                if self.lens2.curv_state == "last_curv" or self.lens2.curv_state == "same" :
                    self.opm.flip(self.lens2surf[0], self.lens2surf[1])

            case "C" | "c":
                # infinity (| (| image
                if self.lens1.curv_state == "last_curv" or self.lens1.curv_state == "same" :
                    self.opm.flip(self.lens1surf[0], self.lens1surf[1])
                if self.lens2.curv_state == "last_curv" or self.lens2.curv_state == "same" :
                    self.opm.flip(self.lens2surf[0], self.lens2surf[1])

            case "D" | "d":
                # infinity (| |) image
                if self.lens1.curv_state == "last_curv" or self.lens1.curv_state == "same" :
                    self.opm.flip(self.lens1surf[0], self.lens1surf[1])
                if self.lens2.curv_state == "first_curv":
                    self.opm.flip(self.lens2surf[0], self.lens2surf[1])

            case "None" | None | "none":
                pass
            case _:
                raise AttributeError('unknown configuration')

        self.update_model()

        # put the right distance between lenses and with the image
        if self.lens2.efl is not None:
            try:
                self.set_lens_gap_efl_affine()
            except:
                self.set_lens_gap_efl_interval()

        self.set_img_dist_bfl()
        self.put_pupil_to_dist(dist_pupil_pair)  # create also self.ffl and self.bfl variables
        self.lenses_gap = self.sm.gaps[self.lens1surf[1]].thi

    ##################################################      CHANGE DISTANCES TO EFL FFL BFL      ##################################################

    def set_lens_gap_efl_interval(self):
        # optimize distance between lenses so efl is the expected focal length
        [a, b, nbboucles] = obtain_distance_to_cross_a_zero(self.opm, self.lens1surf[1], self.expected_fl,
                                                            self.thin_lens_analysis.O1O2)
        distance_between_lenses = optimize.brentq(lens_distance_to_efl_zero, a, b,
                                                  args=(self.opm, self.lens1surf[1], self.expected_fl))

        self.sm.gaps[self.lens1surf[1]].thi = distance_between_lenses

        self.opm.update_model()

    def set_lens_gap_efl_affine(self):
        # usually, the relation between lenses_distance and pair_efl is a kind of 1/x function (Gullstrand formula)
        # we have a relation : 1/efl = a * distance + b
        # we will try to use this to find the necessary lenses separation to have the expected efl

        # initialize 2 distances
        d1 = self.thin_lens_analysis.O1O2
        d2 = d1 + 10

        # get the pair efl for the two distances
        efl_list = [safe_lens_distance_to_efl(d, self.opm, self.lens1surf[1]) for d in [d1,d2]]

        # avoid the singularity point of 1/x
        if efl_list[0] is None :
            d1 = d1 + 20
            efl_list[0] = safe_lens_distance_to_efl(d1, self.opm, self.lens1surf[1])
        if efl_list[1] is None :
            d2 = d2 + 20
            efl_list[1] = safe_lens_distance_to_efl(d2, self.opm, self.lens1surf[1])
        if None in efl_list :
            raise Exception('Error while try to find distance between lenses. Impossible to determine efl for '
                            'two distances')

        # determine affine function coefficients (1/(1/x) = x) look for y=ax+b with y = 1/efl and x = distance
        y1 = 1 / efl_list[0]
        y2 = 1 / efl_list[1]
        a = (y2-y1)/(d2-d1)
        b = -a*d1 + y1

        # determine the theoretical distance to have the correct efl
        defl = (1/self.expected_fl - b)/a
        efl = safe_lens_distance_to_efl(defl, self.opm, self.lens1surf[1])

        if efl is None or (efl - self.expected_fl > 1e-6):
            # if there is an error when check for the EFL with the determined distance
            # or if the obtained efl is not the expected efl
            raise Exception('Impossible to find the distance between lenses to have the expected EFL.')

    def put_pupil_to_dist(self, dist_pupil_pair):
        self.ffl = self.opm['parax_model'].opt_model['analysis_results']['parax_data'].fod.ffl
        self.bfl = self.opm['parax_model'].opt_model['analysis_results']['parax_data'].fod.bfl

        if dist_pupil_pair is None :
            dist = - self.ffl  # FFL front focal length is from first surface to front focal plane
        else :
            dist = dist_pupil_pair

        self.sm.gaps[2].thi = dist

        self.opm.update_model()

    def set_img_dist_bfl(self):
        # distance after the last lens is set to back focal length
        self.sm.gaps[self.lens2surf[1]].thi = self.pm.opt_model['analysis_results']['parax_data'].fod.bfl

        self.opm.update_model()

    ##################################################      VARIOUS STUFF      ##################################################

    def update_model(self, **kwargs):
        self.opm.update_model(**kwargs)
        set_clear_apertures(self.sm, self.avoid_surface_change_list)
        self.opm.update_model(**kwargs)

    def change_field_angle(self, angle):
        if isinstance(angle, list):
            fld = angle
        else:
            fld = [angle]

        self.opm['optical_spec']['fov'] = FieldSpec(self.osp, key=['object', 'angle'], flds=fld, is_relative=False)
        self.update_model()

    def name(self):
        # return in the same str the name of the two files uses in the pair, to be used for print for user after

        name = ""

        if id is not None:
            name = name + "id : " + str(self.id)

        name = name + "   " + os.path.basename(self.lens1.filepath)

        if self.lens2.filepath is not None :
            name = name + " -> " + os.path.basename(self.lens2.filepath)

        if hasattr(self, 'config'):
            name = name + "  config : " + str(self.config)

        return name

    def disp_raytracing(self, do_draw_rays=True, do_paraxial_layout=False, title=None, window_name=None):
        # Draw ray trace for the user
        # if matplotlib figure stay blank, try to add to the code file where is used the function :
        #           import matplotlib
        #           matplotlib.use('TkAgg') #(can need to use other thing in the use function depending on your system)

        fig = plt.figure(FigureClass=InteractiveLayout, opt_model=self.opm,
                          do_draw_rays=do_draw_rays, do_paraxial_layout=do_paraxial_layout).plot()
        if title is None:
            title = 'Ray tracing YZ diagram of pair ' + str(self.short_name)
        fig.suptitle(title)

        if window_name is None:
            window_name = 'YZ diagram pair ' + str(self.short_name)
        fig.canvas.manager.set_window_title(window_name)
        plt.show()
        return fig

    def compute_seidel(self):
        self.seidel = compute_third_order(self.opm)
        return self.seidel

    def decenter_lens(self, lens_to_decenter=1, distance=0, angle=0):
        # decenter the lens 1 or 2 (given lens_to_decenter) by a distance and an angle to the optical axis
        # distance in y direction, and euler angle alpha

        if lens_to_decenter == 1 :
            [idx1, idx2] = self.lens1surf
        elif lens_to_decenter == 2 :
            [idx1, idx2] = self.lens2surf
        else :
            raise ValueError('lens_to_decenter should be 1 or 2')

        # give possibility to decenter surfaces from idx1 to idx2
        if idx1 == idx2 :
            # should not happen but we never know
            self.sm.ifcs[idx1].decenter = srf.DecenterData('dec and return')
        else :
            # surfaces after index 1 are decentered
            self.sm.ifcs[idx1].decenter = srf.DecenterData('decenter')
            # surfaces after index 2 are putted again to previous direction
            self.sm.ifcs[idx2].decenter = srf.DecenterData('reverse')

        # decenter by the distance (in the y direction : cf. RayOptics doc)
        self.sm.ifcs[idx1].decenter.dec[1] = distance
        self.sm.ifcs[idx2].decenter.dec[1] = distance

        # change lens angle (alpha euler angle, cf. RayOptics doc)
        self.sm.ifcs[idx1].decenter.euler[0] = angle
        self.sm.ifcs[idx2].decenter.euler[0] = angle

    def image_radius_one_wvl(self, fld, wvl, attribute=True):
        # determine the maximum radius of an image for the field fld and wavelength wvl after the system
        # attribute bool if True create/replace self.img_radius with value for this computation

        # trace two extremum rays of the pupil
        ray_1 = trace_base(self.opm, [0., 1.], fld, wvl)
        ray_2 = trace_base(self.opm, [0., -1.], fld, wvl)

        max_semidiameter = max(abs(ray_1[0][-1][0][1]), abs(ray_2[0][-1][0][1]))
        # ray_1[0][index surface image][XYZ][Y]

        if attribute:
            self.img_radius = {'radius': max_semidiameter, 'fld': fld, 'wvl': wvl}

        return max_semidiameter

    def compute_image_radius(self, fld=None):
        # determine the maximum radius of an image for the field fld and all wavelengths
        # in case fld=None, the field use of the optical model is the one with the higher y (in absolute value)

        if fld is None:
            fld_list = self.osp.field_of_view.fields  # get fields in the optical model
            if fld_list :  # not empty
                fld = fld_list[0]
                for i in range (1, len(fld_list)):
                    if abs(fld_list[i].y) > abs(fld.y):  # determine the field with the maximum y
                        fld = fld_list[i]
            else:
                return

        wvl_list = self.osp.spectral_region.wavelengths

        if not wvl_list: # empty
            return
        else:
            img_sd = 0
            for wvl in wvl_list :
                img_sd_tmp = self.image_radius_one_wvl(fld, wvl, attribute=False)
                if img_sd_tmp > img_sd :
                    img_sd = img_sd_tmp

        self.img_radius = {'radius': img_sd, 'fld': fld, 'wvl': wvl_list}

        return img_sd


    ##################################################      CUT-OFF ANGLE      ##################################################

    def fld_passing_system(self, fld, wvls=None):
        # check if all rays pass the system
        # take extreme rays and check if they haven't Y out of semidiameter for all interfaces

        if wvls is None :
            wvls = self.osp.spectral_region.wavelengths
        elif type(wvls) is not list:
            wvls = [wvls]

        for wvl in wvls:
            # trace two extremum rays of the pupil
            try:
                f = io.StringIO()
                with contextlib.redirect_stderr(f):
                    # avoid print error
                    ray_1 = trace_base(self.opm, [0., 1.], fld, wvl)
                    ray_2 = trace_base(self.opm, [0., -1.], fld, wvl)
            except Exception as exception:
                # one of rays can't be trace, in general it is too widely out of the system
                return False

            # check for each interface if rays are inside lenses
            for id_surface in range(self.lens1surf[0], self.lens2surf[1]):  # first to last interface of the system
                semidiameter = self.sm.ifcs[id_surface].surface_od()
                max_yray = max(abs(ray_1[0][id_surface][0][1]), abs(ray_2[0][id_surface][0][1]))
                # ray_1[0][id_surface][XYZ][Y]

                if semidiameter < max_yray or np.isnan(max_yray) :
                    return False
        return True

    def determine_cutoffangle_dichotomy(self, angle_start=0, angle_end=45, angle_precision=0.01):
        # only for centered system

        ######  initialize angles
        a1 = angle_start
        a2 = angle_end

        if not 0 <= a1 <= a2 :
            raise ValueError('It is needed that 0 <= angle_start <= angle_end. Please note that '
                             'determine_cutoffangle_dichotomy is designed for centered systems only.')

        def update_error():
            # check if error with the pair optical model
            f = io.StringIO()
            with contextlib.redirect_stderr(f):
                # avoid print error and capture it
                self.update_model()
            return f.getvalue()

        tmp_err = update_error()

        ###### test max and min angles
        # get fields
        # add the field we want to use
        self.osp.field_of_view.fields.append(FieldSpec(self.osp, key=['object', 'angle'], flds=[a1],
                                                       is_relative=False).fields[0])
        # get its index
        i_fld = len(self.osp.field_of_view.fields) - 1

        # update the optical model to adapt the field to the model
        tmp_err = update_error()
        # get the field
        fld = self.osp.field_of_view.fields[i_fld]

        if not self.fld_passing_system(fld):
            # min angle doesn't pass
            self.cutoffangle = {'value' : None, 'angle_start' : angle_start,
                                'angle_end' : angle_end, 'angle_precision' : angle_precision}
            # remove the field used for determining the cut-off angle
            self.osp.field_of_view.fields.pop(i_fld)
            self.update_model()
            return

        elif a1 == a2:
            self.cutoffangle = {'value': angle_start, 'angle_start': angle_start,
                                'angle_end': angle_end, 'angle_precision': angle_precision}
            # remove the field used for determining the cut-off angle
            self.osp.field_of_view.fields.pop(i_fld)
            self.update_model()
            return

        # modify the field with angle a2
        self.osp.field_of_view.fields[i_fld] = FieldSpec(self.osp, key=['object', 'angle'], flds=[a2],
                                                         is_relative=False).fields[0]

        tmp_err = update_error()
        # get the field
        fld = self.osp.field_of_view.fields[i_fld]

        if self.fld_passing_system(fld):
            # the maximal angle to test pass
            self.cutoffangle = {'value' : angle_end, 'angle_start' : angle_start,
                                'angle_end' : angle_end, 'angle_precision' : angle_precision}
            # remove the field used for determining the cut-off angle
            self.osp.field_of_view.fields.pop(i_fld)
            self.update_model()
            return


        ###### dichotomy
        while abs(a2-a1) > angle_precision*0.01:
            # angle to determine
            c = 0.5*(a1+a2)

            # get the field of angle c
            self.osp.field_of_view.fields[i_fld] = FieldSpec(self.osp, key=['object', 'angle'], flds=[c],
                                                             is_relative=False).fields[0]

            tmp_err = update_error()
            fld = self.osp.field_of_view.fields[i_fld]

            # check if fld c is in or out of the system
            if self.fld_passing_system(fld):
                # cut off between c and a2
                a1 = c
            else:
                # cut off between a1 and c
                a2 = c


        ###### end the function and output results
        # remove the field used for determining the cut-off angle
        self.osp.field_of_view.fields.pop(i_fld)
        self.update_model()

        # output the last angle known to succeed passing the system
        # round down depending on precision, to avoid 30 nonsense digits
        # round down to be sure the given output angle pass the system
        value = Decimal(str(a1)).quantize(Decimal(str(angle_precision)), rounding=ROUND_DOWN)

        self.cutoffangle = {'value': value, 'angle_start': angle_start,
                            'angle_end': angle_end, 'angle_precision': angle_precision}

        return

    ##################################################      WAVEFRONT      ##################################################

    def preprocess_wvf(self, fld_list=None, wvl_list=None, foc=0, NbRays=16):
        # this function trace the rays for the described field and wavelength, at given focus
        # fld_list and wvl_list are of type list, foc is float or int, NbRays is int

        if fld_list is None:
            fld_list = self.osp.field_of_view.fields
        if wvl_list is None:
            wvl_list = self.osp.spectral_region.wavelengths

        nb_fld = len(fld_list)
        nb_wvl = len(wvl_list)

        wvf_preprocessed = []
        wvf_processed = []

        for i, fld in enumerate(fld_list):
            wvl_wvf = []
            wvl_wvf_processed = []

            ref_fld = fld
            ref_wvl = wvl_list[0]

            for j, wvl in enumerate(wvl_list):
                # preprocess the wavefront at field fld, wavelength wvl and focus foc, with NbRays rays
                grid, upd_grid, ray00_pkg, pre_piston = ta.trace_wavefront(self.opm, fld, wvl, foc, ref_fld, ref_wvl,
                                                                           num_rays=NbRays, return_piston=True)

                wvl_wvf.append({'wvl': wvl, 'wvl_number': j, 'foc': foc, 'grid': grid, 'upd_grid': upd_grid,
                                'ray00_pkg': ray00_pkg, 'pre_piston': pre_piston})
                wvl_wvf_processed.append({'wvl': wvl, 'wvl_number': j, 'foc': foc, 'wvf_grid': None})

            wvf_preprocessed.append({'fld': fld, 'fld_number': i, 'wvl_and_data': wvl_wvf})
            wvf_processed.append({'fld': fld, 'fld_number': i, 'wvl_and_data': wvl_wvf_processed})

        self.wvf_preprocessed = wvf_preprocessed
        self.wvf_processed = wvf_processed

    def get_number_list_from_fld_wvl(self, fld_list=None, wvl_list=None):
        if fld_list is None:
            fld_number_list = [i for i in range(len(self.wvf_preprocessed))]
        else:
            fld_number_list = []
            nb_fld_preprocessed = len(self.wvf_preprocessed)

            for fld in fld_list:
                for i in range(nb_fld_preprocessed):
                    if self.wvf_preprocessed[i]['fld'] == fld:
                        fld_number_list.append(i)

        # get the index of every wvl in self.wvf_preprocessed
        if wvl_list is None:
            wvl_number_list = [i for i in range(len(self.wvf_preprocessed[0]['wvl_and_data']))]
        else:
            wvl_number_list = []

            if len(self.wvf_preprocessed) != 0:
                nb_wvl_preprocessed = len(self.wvf_preprocessed[0]['wvl_and_data'])
            else:
                nb_wvl_preprocessed = 0

            for wvl in wvl_list:
                for i in range(nb_wvl_preprocessed):
                    if self.wvf_preprocessed[0]['wvl_and_data'][i] == wvl:
                        wvl_number_list.append(i)

        if not (fld_number_list and wvl_number_list) :
            # if fld list empty and wvl list empty
            raise AttributeError('unknown fld and wvl for wavefront refocusing computation')
        elif not fld_number_list :
            # fld_number_list empty
            raise AttributeError('unknown fld for wavefront refocusing computation')
        elif not wvl_number_list :
            # wvl_number_list empty
            raise AttributeError('unknown wvl for wavefront refocusing computation')

        return [fld_number_list, wvl_number_list]

    def refocus_wvf_fld_wvl(self, fld_list=None, wvl_list=None, foc=0):
        # refocus the wavefront at the desired distance
        # input are the field fld instance and wavelength in nm

        # get the index of every fld in self.wvf_preprocessed
        [fld_number_list, wvl_number_list] = self.get_number_list_from_fld_wvl(fld_list, wvl_list)

        self.refocus_wvf_number(fld_number_list, wvl_number_list, foc)

    def refocus_wvf_number(self, fld_number_list=None, wvl_number_list=None, foc=0):
        # refocus the wavefront at the desired distance

        if fld_number_list is None:
            fld_number_list = [i for i in range(len(self.wvf_preprocessed))]
        if wvl_number_list is None:
            wvl_number_list = [i for i in range(len(self.wvf_preprocessed[0]['wvl_and_data']))]

        for i in fld_number_list:

            ref_fld = self.wvf_preprocessed[i]['fld']
            ref_wvl = self.wvf_preprocessed[i]['wvl_and_data'][0]['wvl']

            for j in wvl_number_list:
                grid_pkg = [self.wvf_preprocessed[i]['wvl_and_data'][j]['grid'],
                            self.wvf_preprocessed[i]['wvl_and_data'][j]['upd_grid']]
                piston_pkg = [self.wvf_preprocessed[i]['wvl_and_data'][j]['ray00_pkg'],
                              self.wvf_preprocessed[i]['wvl_and_data'][j]['pre_piston']]
                fld, wvl = self.wvf_preprocessed[i]['fld'], self.wvf_preprocessed[i]['wvl_and_data'][j]['wvl']

                # refocused grid
                self.wvf_processed[i]['wvl_and_data'][j]['wvf_grid'] = ta.focus_wavefront(self.opm, grid_pkg,
                                                                                          fld, wvl, foc,
                                                                                          ref_fld, ref_wvl,
                                                                                          piston_pkg=piston_pkg,
                                                                                          image_pt_2d=None,
                                                                                          image_delta=None,
                                                                                          value_if_none=np.nan)[:,:, 2]


    def compute_rms_wvf_no_weight(self, fld_number_list=None, wvl_number_list=None, add_rms_entry=False,
                                  polychromatic='rmse_rms'):
        # if add_rms_entry is True, add a variable to rms_detail.
        # If False, erase and replace rms_detail with the new results.

        if fld_number_list is None:
            fld_number_list = [i for i in range(len(self.wvf_preprocessed))]
        if wvl_number_list is None:
            wvl_number_list = [i for i in range(len(self.wvf_preprocessed[0]['wvl_and_data']))]

        # for one field, one wavefront, RMS is defined as the standard deviation of the wavefront error map
        #
        # for several field, several wavefront, took the RMS of RMS wavefront error of each fld, wvl
        #
        # fld and wvl weight are ignored and all considered as equal

        nb_fld = np.max(np.shape(fld_number_list))
        nb_wvl = np.max(np.shape(wvl_number_list))
        nb_px_side_wvf = np.max(
            np.shape(self.wvf_processed[fld_number_list[0]]['wvl_and_data'][wvl_number_list[0]]['wvf_grid']))

        if nb_fld > 1 or nb_wvl > 1 :
            p_v = None
            match polychromatic:
                case 'rmse_rms' | 'worst':
                    rms_individual = []
                    for i_fld in fld_number_list :
                        for i_wvl in wvl_number_list :
                            wvf_tmp = self.wvf_processed[fld_number_list[i_fld]]['wvl_and_data'][wvl_number_list[i_wvl]]['wvf_grid']
                            wvf_tmp_mask = np.logical_not(np.isnan(wvf_tmp))
                            rms_individual.append(wvf_tmp[wvf_tmp_mask].std())

                    wvf_concatenate = None

                    if polychromatic == 'rmse_rms':
                        rms = np.sqrt(np.sum(np.square(rms_individual))/len(rms_individual))
                    elif polychromatic == 'worst':
                        rms = np.max(rms_individual)

                    else :
                        raise ValueError("polychromatic should be 'rmse_rms' or 'worst'")

                case _ :
                    raise ValueError("polychromatic should be 'rmse_rms' or 'worst'")

        else:
            # RMS one field one wavelength
            wvf_concatenate = self.wvf_processed[fld_number_list[0]]['wvl_and_data'][wvl_number_list[0]]['wvf_grid']
            rms = wvf_concatenate[np.logical_not(np.isnan(wvf_concatenate))].std()  # RMSE = standard dev
            try :
                p_v = (wvf_concatenate[np.logical_not(np.isnan(wvf_concatenate))].max() -
                       wvf_concatenate[np.logical_not(np.isnan(wvf_concatenate))].min())
            except :
                p_v = None

        fld_list = [self.wvf_preprocessed[j]['fld'] for j in fld_number_list]
        wvl_list = [self.wvf_preprocessed[0]['wvl_and_data'][j]['wvl'] for j in wvl_number_list]

        if len(fld_number_list)>1 or len(wvl_number_list)>1 :
            polychromatic_tmp = polychromatic
        else :
            polychromatic_tmp = 'none'

        if not (hasattr(self, 'rms_detail') and add_rms_entry):
            self.rms_detail = [{'RMS': rms, 'fld_number_list': fld_number_list, 'wvl_number_list': wvl_number_list,
                                'used_fld': fld_list, 'used_wvl': wvl_list, 'polychromatic': polychromatic_tmp,
                                'P-V': p_v,
                                'foc':self.wvf_processed[fld_number_list[0]]['wvl_and_data'][wvl_number_list[0]]['foc'],
                                'grid size': len(self.wvf_processed[0]['wvl_and_data'][0]['wvf_grid'])}]
        else:
            self.rms_detail.append({'RMS': rms, 'fld_number_list': fld_number_list, 'wvl_number_list': wvl_number_list,
                                    'used_fld': fld_list, 'used_wvl': wvl_list, 'polychromatic': polychromatic_tmp,
                                    'P-V': p_v,
                                    'foc':self.wvf_processed[fld_number_list[0]]['wvl_and_data'][wvl_number_list[0]]['foc'],
                                    'grid size': len(self.wvf_processed[0]['wvl_and_data'][0]['wvf_grid'])})

        return rms, fld_number_list, wvl_number_list, wvf_concatenate

    def delete_wvf(self, delete_preprocessed_wavefront=True, delete_processed_wavefront=True):
        # release memory of wavefront calculations, delete just the grids
        if delete_preprocessed_wavefront:
            for fld in self.wvf_preprocessed:
                for wvl in fld['wvl_and_data']:
                    # suppress the variable
                    del wvl['grid']
                    del wvl['upd_grid']

        if delete_processed_wavefront:
            for fld in self.wvf_processed:
                for wvl in fld['wvl_and_data']:
                    # suppress the variable
                    del wvl['wvf_grid']

        # force to free unreferenced memory
        gc.collect()

    def delete_wvf_total(self, delete_preprocessed_wavefront=True, delete_processed_wavefront=True):
        # release memory of wavefront calculations
        if delete_preprocessed_wavefront:
            # suppress the variable
            del self.wvf_preprocessed
        if delete_processed_wavefront:
            del self.wvf_processed

        # force to free unreferenced memory
        gc.collect()

    def refocus_compute_rms_for_optimize(self, foc, fld_number_list=None, wvl_number_list=None,
                                         polychromatic='rmse_rms'):
        self.refocus_wvf_number(fld_number_list, wvl_number_list, foc)
        rms, _, _, _ = self.compute_rms_wvf_no_weight(fld_number_list, wvl_number_list, polychromatic=polychromatic)
        return rms

    def best_rms_focus(self, fld_number_list=None, wvl_number_list=None, start_foc=0, focus_polychromatic='ref_wvl_i'):
        # this function optimize the focus to have the minimum RMS
        # it is better to have preprocessed the wavefront before, with self.wvf_preprocessed() (in order to have the
        # desire precision, fld, etc)
        # focus_polychromatic = 'ref_wvl_i', 'rmse_rms'

        # if a preprocessed wavefront instance doesn't exist yet, determine it
        if not hasattr(self, 'wvf_preprocessed'):
            self.preprocess_wvf()

            # if no specific fld and wvl is chosen, get them all
            # if no specific fld and wvl is chosen, reference fld, wvl is the fld 0 wvl 0
        if fld_number_list is None:
            fld_number_list = [0]
            # fld_number_list = [i for i in range(len(self.wvf_preprocessed))]
        if wvl_number_list is None:
            if focus_polychromatic == 'ref_wvl_i':
                wvl_number_list = [0]
                # self.reference_wvl
            else:
                wvl_number_list = [i for i in range(len(self.wvf_preprocessed[0]['wvl_and_data']))]

        # optimize the focus to have the minimum RMS
        foc_opti = optimize.minimize(self.refocus_compute_rms_for_optimize, start_foc,
                                     args=(fld_number_list, wvl_number_list, focus_polychromatic),
                                     method="BFGS", options={'gtol': 1e-3})
        # foc_opti.success said if successed to find a minimum
        # foc_opti.x = foc to have the minimum

        self.foc_opti = foc_opti

        return foc_opti

    def disp_wavefront(self, fld_number_list=None, wvl_number_list=None, cyclic=True, subplot=True):
        # plot plt schematics of the wavefront
        # cyclic : the wavefront will be displayed between 0 to 1 wave unit
        # subplot : all wavefront in one single figure

        if fld_number_list is None:
            fld_number_list = [i for i in range(len(self.wvf_preprocessed))]
        if wvl_number_list is None:
            wvl_number_list = [i for i in range(len(self.wvf_preprocessed[0]['wvl_and_data']))]

        nb_fld = len(fld_number_list)
        nb_wvl = len(wvl_number_list)

        if nb_wvl<1 or nb_fld<1:
            # no field, no wavelength
            return

        subplot = subplot and (nb_fld + nb_wvl - 1 > 1)

        if cyclic :
            title_global = "Wavefront error map (modulo 1 wave)"
        else:
            title_global = "Wavefront error map, not cyclic"

        if subplot:
            fig, axs = plt.subplots(nb_fld, nb_wvl)
            fig.suptitle(title_global)
            fig.canvas.manager.set_window_title('Wavefront error map pair ' + str(self.short_name))

        for i_subplot, i in enumerate(fld_number_list):
            for j_subplot, j in enumerate(wvl_number_list):
                # get the wavefront for fld[i] and wavelength[j]
                wvf = self.wvf_processed[i]['wvl_and_data'][j]['wvf_grid']

                # avoid NaN values (which aren't part of the wavefront map)
                wvf_mask = np.ma.masked_where(np.logical_not(np.isnan(wvf)), wvf)

                if cyclic:
                    # modulo 1 wave
                    wvf_mask = np.mod(wvf_mask, 1)

                # get fld and wvl for which the wavefront will be displayed
                fld = self.wvf_preprocessed[i]['fld']
                wvl = self.wvf_preprocessed[i]['wvl_and_data'][j]['wvl']

                # create the figure
                if subplot:
                    if nb_fld>1 and nb_wvl>1:
                        ax = axs[i_subplot, j_subplot]
                    elif nb_fld==1:
                        ax = axs[j_subplot]
                    elif nb_wvl==1:
                        ax = axs[i_subplot]
                    else:
                        raise ValueError('there is no field and no wavelength')
                else:
                    fig = plt.figure()
                    ax = plt.gca()
                    fig.canvas.manager.set_window_title('Wavefront error map pair ' + str(self.short_name))

                if cyclic:
                    im = ax.imshow(wvf_mask.data, cmap='hsv', vmin=0, vmax=1)
                else:
                    im = ax.imshow(wvf_mask.data, cmap='jet')

                if subplot :
                    title_start = ''
                else:
                    title_start = title_global + '\n'

                title = (title_start + "Pair : " + str(self.short_name) + "\nwvl : " + str(wvl) + " nm     fld : " +
                         "{:.1f}".format(fld.x) + " x, " + "{:.1f}".format(fld.y) + " y")
                ax.set_title(title, fontsize=10)
                ax.grid(visible=False)

                if subplot:
                    cbar = plt.colorbar(im, ax=ax)
                    cbar.set_label('waves', rotation=270)
                else:
                    cbar = plt.colorbar(im, ax=ax)
                    cbar.set_label('waves', rotation=270, fontsize=12, labelpad=15)
        plt.show()



# select_pair create pair array, then for each pair check distances and determine RMS, in multiprocessing
# some information are keep in a pandas dataframe and which is returned by each core
# best pairs are created again to check their RMS wavefront error with a better precision

# import from general libraries

# optical materials information, to initialize subprocess well
from opticalglass.schott import SchottGlass
from opticalglass.ohara import OharaGlass
from opticalglass.cdgm import CDGMGlass
from opticalglass.hikari import HikariGlass
from opticalglass.hoya import HoyaGlass
from opticalglass.sumita import SumitaGlass

from rayoptics.environment import *  # to import RayOptics
from copy import deepcopy  # to copy an object
import logging
import multiprocess as mp
import os.path  # to get filename form path
import sys
from functools import partial  # create function of one variable from function of several variables
from itertools import chain
import time
from datetime import datetime
import psutil  # force a subprocess to execute on a specific core
from ctypes import c_int  # to be able to create a safe counter for multiprocessing
import numpy as np
import pandas as pd

# import from PyRFTL
from pyrftl.func.various_stuff import is_between, time_str_hms_from_s, set_log, description_pair_list, pair_df_2_lens_config
from pyrftl.func.pairs import Pair
from pyrftl.func.wavefront_pair_analysis import wvf_fast_one_pair
from pyrftl.func.distances_pair_analysis import check_pair_thin_lens, check_pair_ray_tracing
from pyrftl.func.lens_selection import open_lenses
import pyrftl.func.save as save
from pyrftl.func.multiprocessing_tools import create_proc_idx_queue, log_process

logger = logging.getLogger('PyRFTL')
# log level
LOG_DEBUG = 10
LOG_INFO = 20
LOG_WARNING = 30
LOG_ERROR = 40
LOG_CRITICAL = 50


def select_pair(param) :
    # select pairs of lenses which correspond to good distance, and determine their wavefront RMS error
    # first open lenses files, then for each pair of lenses (and configurations, so lenses directions), check for
    # distances and if it is ok, determine the RMS. Save values in a pandas dataframe

    t00 = time.time()

    if hasattr(param, 'start_time'):
        start_date = datetime.fromtimestamp(param['start_time']).strftime("%Y_%m_%d_%Hh%M")
    else:
        start_date = datetime.fromtimestamp(t00).strftime("%Y_%m_%d_%Hh%M")
    print('Start analysis time : ' + start_date)

    # set logger
    set_log(param['save_folder_path'], False)

    logger = logging.getLogger('PyRFTL')  # log

    logger.info('Start analysis : ' + param['analysis_name'])
    logger.info('date and time at log start: ' + start_date)

    logger.info('Parameters for analysis : ' + str(param))

    if param['dist_criteria']['O1O2_min'] is None or param['dist_criteria']['O1O2_min']<0 :
        param['dist_criteria']['O1O2_min'] = 0
        # lenses cannot be nearest than 0 mm, it is physically impossible

    # create a list of lenses to use
    lenses_list = open_lenses(param['lenses_param'])

    # create folders to save seq model and roa
    if param['save_seq_detail'] != 'no':
        dir_path = os.path.join(param['save_folder_path'], 'Pairs_Sequential_Model_detail')
        os.mkdir(dir_path)
    if param['save_roa'] != 'no':
        dir_path = os.path.join(param['save_folder_path'], 'RayOptics models')
        os.mkdir(dir_path)

    # create pairs of lenses and select good ones
    # config should be a list
    if not isinstance(param['config'], list):
        param['config'] = [param['config']]

    # select lenses
    # first selection
    pair_data_list, pair_dict = which_pair_to_add_successive(lenses_list, param)

    # check if at least one pair was selected
    if len(pair_data_list) <= 1 :  # only column label
        logger.info("No pair have been selected !\nYou should add more lenses or change your criteria.")
        print("No pair have been selected !\nYou should add more lenses or change your criteria.")
        tff = time.time()
        logger.info('Total time : ' + time_str_hms_from_s(tff - t00))
        print(time_str_hms_from_s(tff - t00))
        return None, pair_dict, None

    # determine pairs with best rms
    pair_dataframe = pd.DataFrame(pair_data_list[1:], columns=pair_data_list[0])
    pair_dataframe.sort_values(by='first analysis RMS wavefront error', inplace=True)  # sort by RMS
    pair_dataframe.set_index('pair short name', inplace=True)  # use pair short name as row index
    # (pair short name should be unique, it is lens1id_lens2id_config)

    # get RMS better resolution
    if param['nbRays_high_comparison'] > param['nbRays_first_comparison']:

        # get rms for best pairs
        rms_best_pairs(pair_dataframe, pair_dict, param, lenses_list)

        # sort dataframe by rms
        pair_dataframe.sort_values(by=['second analysis RMS wavefront error', 'first analysis RMS wavefront error'],
                                   inplace=True)

    # check if at least one pair as a better rms than diffraction
    if not pair_dataframe.iloc[0].at['<lambda/14'] :
        logger.info('No pair is good enough. All have more wavefront aberrations than the diffraction limit '
                    '(RMS error > lambda/14).\nYou should try to add more lenses.\n')
    else:
        logger.info('best pair is : ' + str(pair_dataframe.index[0]) + '\n')

    # export data
    save.save_results_csv_df(param, pair_dataframe)

    tff = time.time()
    logger.info('Total time : ' + time_str_hms_from_s(tff-t00))
    print(time_str_hms_from_s(tff - t00))

    if param['save_ram'] == 'no':
        # no need to keep lenses if all pairs model are exported and do not need to be recreated from lenses
        lenses_list_export = None
    else:
        lenses_list_export = lenses_list

    return pair_dataframe, pair_dict, lenses_list_export


def init_subprocess (q=None, queue_for_log=None, counter_safe=None, log_mute=True) :
    # initialize subprocess to force to execute it on a specific core. Also initialize other stuff
    # q = multiprocessing queue with core number to use for each subprocess
    # q = None if no need to force subprocess on a specific core

    # mute logging in the subprocess. In the future, should redirect it in queue_log as a handler. For now, just mute
    # others logs than mine
    if log_mute :
        logging.basicConfig(handlers=[logging.StreamHandler(sys.stdout)],level=60)

    # force the subprocess to execute on a specific core. The core number is taken from queue q
    if q is not None :
        i_proc = q.get()
        psutil.Process().cpu_affinity([i_proc])

    # initiate the queue in which logs will be sent
    global queue_log
    queue_log = queue_for_log

    # initiate the shared variable (between subprocesses) to observe/count how many pairs have been processed
    global counter
    counter = counter_safe

    # initialize glass catalogs
    # if not, medium instances of glass type will have None as catalog attribute in subprocesses
    SchottGlass.initialize_catalog(None)
    SumitaGlass.initialize_catalog(None)
    OharaGlass.initialize_catalog(None)
    HikariGlass.initialize_catalog(None)
    HoyaGlass.initialize_catalog(None)
    CDGMGlass.initialize_catalog(None)


def which_pair_to_add_successive(lenses_list, param):
    # return a dataframe with data of pairs which succeed in distance test.
    # Can return also a list of pair instances (optional)

    nb_lenses = len(lenses_list)

    # initialization of multiprocessing
    # create list of possible lenses combinations to evaluate each one separately via multiprocessing
    lenses_possible_pairs = []
    if ('pair_of_one_lens' in param) and param['pair_of_one_lens']:  # if 'pair_of_one_lens' exist and is True
        lenses_possible_pairs = [(lens, None) for lens in lenses_list]
    else:
        for i in range(nb_lenses):
            for j in range(nb_lenses):
                lenses_possible_pairs.append((lenses_list[i], lenses_list[j]))

    # create queue used to initialize processor cores
    q = create_proc_idx_queue(param)

    queue_log = mp.Queue()  # each element of the queue is an array of two elements : [message, log level]
    counter = mp.Value(c_int)

    # start process to collect logs and write them into a log file
    process_log = mp.Process(target=log_process, args=(queue_log, param))
    process_log.start()

    # start processes to do calculations
    with mp.Pool(param['nb_subprocesses'], initializer=init_subprocess, initargs=(q, queue_log, counter)) as pool:
        t0 = time.time()

        # get list of valid pairs with multiprocessing.
        output = pool.map(partial(which_config_to_add_lens_successive, param=param,
                                  nb_without_config=nb_lenses**2), lenses_possible_pairs)

        print()

        # extract list of data and list of pairs (pair list is empty if in other save ram mode than 'no')
        pair_data_list = description_pair_list(param)  # get columns names
        pair_dict = {}  # keys are pairs short names
        for elt in output:
            pair_data_list = pair_data_list + elt[0]  # concatenate list
            pair_dict = {**pair_dict, **elt[1]}  # concatenate dictionary

    # When multiprocessing workers from pool are finished, put None in log_queue to ask process_log to stop
    queue_log.put(None)
    process_log.join()

    return pair_data_list, pair_dict


def rms_best_pairs(pair_dataframe, pair_dict, param, lenses_list):
    # determine rms wavefront error for best pairs (given first rms determination criteria)
    # pair_dataframe should already be sorted with first row = best pairs

    # get first rows of the dataframe
    pair_dataframe_best = pair_dataframe.head(param['number_best_pair_high_comparison'])

    # remove rows with RMS lower than criteria (with tolerance in %)
    pair_dataframe_best = pair_dataframe_best[pair_dataframe_best['first analysis RMS wavefront error'] <
                                              1 / 14 * (1 + param['tolerance_wavefront_RMS_high_percent'] * 0.01)]

    # create list of lenses pairs to use
    lenses_pairs_best = []
    for linekey in pair_dataframe_best.index:
        # for each line of dataframe, took lens combination of the pair, or the pair
        if param['save_ram'] != 'no':
            config, lens1, lens2 = pair_df_2_lens_config(pair_dataframe_best, linekey, lenses_list)
            lenses_pairs_best.append([config, lens1, lens2])
        else:
            lenses_pairs_best.append(pair_dict[linekey])  # linekey is pair short name

    nb_pairs = len(lenses_pairs_best)

    # create queue used to initialize processor cores
    q = create_proc_idx_queue(param)

    queue_log = mp.Queue()  # each element of the queue is an array of two elements : [message, log level]
    counter = mp.Value(c_int)

    # start process to collect logs and write them into a log file
    process_log = mp.Process(target=log_process, args=(queue_log, param))
    process_log.start()

    # start processes to do calculations
    with mp.Pool(param['nb_subprocesses'], initializer=init_subprocess,
                 initargs=(q, queue_log, counter)) as pool:
        t0 = time.time()

        print("Start computing RMS for best pairs\n"
              "Processed 0/" + str(nb_pairs) + " pairs", end="")

        # get list of valid pairs with multiprocessing. Use not blocking function.
        output = pool.map(partial(wavefront_rms_pair_or_lenses, param=param, nb_pairs=nb_pairs), lenses_pairs_best)

        print()

    # When multiprocessing workers from pool are finished, put None in log_queue to ask process_log to stop
    queue_log.put(None)
    process_log.join()

    # save results
    for i in range(len(output)):
        [short_name, rms, bfl_opti, pair] = output[i]

        if pair is not None :
            pair_dict[short_name] = pair

        try :
            if pair_dataframe.index[i] != short_name:
                # check if we modify the right line
                raise ValueError('Pair has changed its position in the dataframe, pair name : ' + short_name +
                                 ' index at its position : ' + str(pair_dataframe.index[i]))

            pair_dataframe.at[short_name, 'optimum bfl for RMS wavefront'] = bfl_opti
            pair_dataframe.at[short_name, 'second analysis RMS wavefront error'] = rms
            pair_dataframe.at[short_name, '<lambda/14'] = rms < 1/14

        except Exception as exception:
            logger.error(exception)


def wavefront_rms_pair_or_lenses(pair_indications, param, nb_pairs = None):
    # determine RMS with huge grid, to determine it only with interesting pairs
    # can either work with an already existing pair, or (re)create it from few information

    if isinstance(pair_indications, Pair):
        pair = pair_indications
    else:  # (re)create the pair
        config, lens1, lens2 = pair_indications
        pair = Pair(lens1, lens2, param['expected_focal_length'], None,
                    {'lens1 id :', lens1.id, 'lens2 id :', lens2.id})
        pair.create_optical_model(config=config, wvl_and_weight=param['wvl_and_weight'], pupil=param['pupil'], fld=[0],
                                  dist_pupil_pair=param['dist_pupil_pair'])

    # compute RMS wavefront error
    wvf_fast_one_pair(pair, param, param['nbRays_high_comparison'])
    rms = pair.rms_detail[0]['RMS']  # get the rms
    bfl_opti = pair.bfl + pair.foc_opti.x[0]  # get focus distance to have the better RMS

    # save seq model details in txt file
    if param['save_seq_detail'] != 'no':
        dir_path = os.path.join(param['save_folder_path'], 'Pairs_Sequential_Model_detail')
        try:
            filename = param['analysis_name'] + "__" + pair.short_name + "_SeqMod"
            save.export_detail_pair_seqmod_txt(pair, param, dir_path=dir_path, overwrite_file=True, filename=filename)
        except Exception as exception:
            logger.error('unable to export sequential model details of pair ' + pair.name() +
                         '\nerror is : ' + str(exception))

    # save roa model in file
    if param['save_roa'] != 'no':
        dir_path = os.path.join(param['save_folder_path'], 'RayOptics models')
        try:
            filename = param['analysis_name'] + "__" + pair.short_name + "_RayOpticsModel"
            pair.opm.save_model(os.path.join(dir_path, filename))
        except Exception as exception:
            logger.error('unable to save rayoptics model of pair ' + pair.name() +
                         '\nerror is : ' + str(exception))

    # do not output best pairs if want to save the most memory as possible
    if False :
        pair = None

    with counter.get_lock():
        counter.value = counter.value + 1
        print('\rProcessed ' + str(counter.value) + "/" + str(nb_pairs) + " pairs", end="")

    return [pair.short_name, rms, bfl_opti, pair]


def which_config_to_add_lens_successive(lens12_tuple, param, nb_without_config=None):
    # get both lenses
    (lens1, lens2) = lens12_tuple

    # create a pair
    if param['save_ram'] == "max":
        copy_lens = False
    else:
        copy_lens = True
    pair = Pair(lens1, lens2, param['expected_focal_length'], None, {'lens1 id :', lens1.id, 'lens2 id :', lens2.id},
                copy_lens)

    # initialisation
    add_pair = True
    pair_dict_tmp = {}

    ####### THIN LENS ANALYSIS #######
    # check with thin lens analysis if pair met criteria
    if param['thin_lens_first_analysis']:
        if not check_pair_thin_lens(pair, param['dist_criteria']):
            add_pair = False
            why_not_add_pair = ("Pair is rejected by thin lens analysis, it does not meet distances criteria. You can "
                                "change them if needed.")

    ####### GEOMETRICAL RAY TRACING ANALYSIS #######
    # check with geometrical ray tracing if pair met criteria

    if add_pair:
        # we can have as many pairs to test as configs we wanted to check
        for pair_config in param['config']:
            try:
                pair_copy = deepcopy(pair)

                # create optical model
                fld0 = [0]
                pair_copy.create_optical_model(config=pair_config, wvl_and_weight=param['wvl_and_weight'],
                                               pupil=param['pupil'], fld=fld0,
                                               dist_pupil_pair=param['dist_pupil_pair'])

                # check distance criteria
                add_pair_config, why_not_add_pair = check_pair_ray_tracing(pair_copy, param)

                if add_pair_config:
                    if ('image_size' in param) and param['image_size']:
                        try:
                            pair_copy.osp['fov'] = FieldSpec(pair_copy.osp, key=['object', 'angle'],
                                                             flds=[param['cut_off_angle_min']], is_relative=False)
                            pair_copy.opm.update_model()
                            pair_copy.compute_image_radius()
                            pair_copy.osp['fov'] = FieldSpec(pair_copy.osp, key=['object', 'angle'],
                                                             flds=fld0, is_relative=False)
                            pair_copy.opm.update_model()

                        except Exception as exception:
                            pass

                    pair_dict_tmp[pair_copy.short_name]=pair_copy

                    if queue_log is not None:
                        queue_log.put(["Pair selected for RMS Wavefront comparison : " + pair_copy.name() + "\n",
                                       LOG_INFO])

                else:
                    if queue_log is not None:
                        queue_log.put(["Pair has been rejected :" + pair_copy.name() + "\n" + why_not_add_pair + "\n",
                                       LOG_INFO])

            except Exception as exception:
                if queue_log is not None:
                    queue_log.put(["Error in which_pair_to_add_lens\n" +
                                   "Error while working on pair : " + pair.name() + "\n" +
                                   "Error is : " + str(exception),
                                   LOG_ERROR])

    else:  # pair does not pass thin lens distance analysis
        if queue_log is not None:
            queue_log.put(["Pair has been rejected : " + pair.name() + "\n" + why_not_add_pair + "\n",
                           LOG_INFO])

    ####### RMS ANALYSIS #######
    for pair_key in pair_dict_tmp:
        pair = pair_dict_tmp[pair_key]
        wvf_fast_one_pair(pair, param, param['nbRays_first_comparison'], queue_log=queue_log)

    pair_description_list = description_pair_list(param, pair=pair_dict_tmp, columns_names=False)

    if param['save_ram'] != "no":
        pair_dict_output = {}
    else :
        pair_dict_output = pair_dict_tmp

    # save seq model details in txt file
    if param['save_seq_detail'] == 'all':
        dir_path = os.path.join(param['save_folder_path'], 'Pairs_Sequential_Model_detail')
        filename = param['analysis_name'] + "__" + pair.short_name + "_SeqMod"
        for pair_key in pair_dict_tmp:
            pair = pair_dict_tmp[pair_key]
            try:
                save.export_detail_pair_seqmod_txt(pair, param, dir_path=dir_path, filename=filename)
            except Exception as exception:
                logger.error('unable to export sequential model details of pair ' + pair.name() +
                             '\nerror is : ' + str(exception))

    # save roa model in file
    if param['save_roa'] == 'all':
        dir_path = os.path.join(param['save_folder_path'], 'RayOptics models')
        for pair_key in pair_dict_tmp:
            pair = pair_dict_tmp[pair_key]
            try:
                filename = param['analysis_name'] + "__" + pair.short_name + "_RayOpticsModel"
                pair.opm.save_model(os.path.join(dir_path, filename))
            except Exception as exception:
                logger.error('unable to save rayoptics model of pair ' + pair.name() +
                             '\nerror is : ' + str(exception))

    # inform user
    try:
        with counter.get_lock():
            counter.value = counter.value + 1
            if counter.value == 1:
                print('Selection of pairs in process...\n0', end="")
            elif nb_without_config is not None:
                print('\r ' + str(round(counter.value / nb_without_config * 100)) + ' %', end="")
                if nb_without_config == counter.value :
                    print('\nCollecting data from threads, please wait...', end="")
            else:
                print('\r ' + str(counter.value), end="")
    except:
        pass

    # return results
    return pair_description_list, pair_dict_output





# import from general libraries
from rayoptics.environment import *  # to import RayOptics
import numpy as np


class Lens:
    # This class is intended to be composed of one lens element. Its unit should be in [mm]
    # Typically, it will be used to open file that contain a lens : Zemax .zmx, CodeV .seq or RayOptics .roa
    # zmx, seq and roa can be in mm, cm or inch unit.
    #
    # Some of its attributes could be :
    # the path of the file lens : filepath and its original system units : original_unit
    # Optical model : opm, sm, osp, pm, em, pt, ar
    # idx = [index of the first surface of the lens, index of the last surface of the lens]
    # air_gaps => True/False if some air layers exist before/after the lens in the original file
    #            None if it was not tested
    # delete_air_gap => True/False if it was ask the air gaps to be deleted.
    #               If air_gaps is false, delete_air_gap is nonsense. If delete_air_gap is None, air_gaps was not tested
    # efl = effective focal length of the lens
    # diameter_minmax = [min diameter of the lens, max diameter of the lens] (diameter = 2*semi-diameter)
    # curv = [curvature of the first surface of the lens, curvature of the last surface of the lens]
    # curv_state = "first_curv" "last_curv" "same" => respectively first and last surface of the lens is the more curved
    #              "same" is when curve of both surface is equal

    def __init__(self, filepath, check_air_gap=True, delete_air_gap=False, id_lens=None, info=None):
        # filepath : path of the Lens file
        # check_air_gap : check if there is air gap before or after the lens, and if delete_air_gap : delete them
        self.filepath = filepath
        self.id = id_lens
        self.info = info

        if filepath is not None:
            # open the lens and get its optical model
            self.opm = open_model(filepath, post_process_imports=False)  # function of rayoptics

            # some useful variables
            self.sm = self.opm['seq_model']
            self.osp = self.opm['optical_spec']
            self.pm = self.opm['parax_model']
            self.em = self.opm['ele_model']
            self.pt = self.opm['part_tree']
            self.ar = self.opm['analysis_results']

            # unit of the lens file ([mm], [cm], [inches])
            self.original_unit = self.opm.system_spec.dimensions

            # convert to mm
            if self.original_unit == "inches":
                self.sm.apply_scale_factor(1 / 25.4)
                self.opm.system_spec.dimensions = "mm"
                self.opm.udpate_model()

            elif self.original_unit == "cm":
                self.sm.apply_scale_factor(0.1)
                self.opm.system_spec.dimensions = "mm"
                self.opm.udpate_model()

            elif self.original_unit != "mm":
                raise ValueError('Error : unknown file unit. Possible file units are inches, mm and cm.')

            # lens interface index
            # [first lens interface, last lens interface]
            if self.sm.get_num_surfaces() > 2:
                self.idx = [1, self.sm.get_num_surfaces() - 2]
                # -2 because python index start at 0, not 1, and last surface is the image
            else:
                raise Exception('No lens in file ')

            # check if there is air gap before or after the lens
            if check_air_gap:
                self.air_gaps = self.check_for_air(delete_air_gap, self.idx)
                self.delete_air_gap = delete_air_gap
            else:
                self.air_gaps = None
                self.delete_air_gap = None

            self.opm.update_model()

            # get the effective focal length
            self.efl = self.opm['parax_model'].opt_model['analysis_results']['parax_data'].fod.efl

            # get the lens min and max diameters
            self.diameter_minmax = self.get_diameter()

            # get lens extremities curvatures
            self.opm.radius_mode = False  # go for curvature mode
            self.opm.update_model()
            self.curv = [self.sm.ifcs[self.idx[0]].profile_cv,
                         self.sm.ifcs[self.idx[1]].profile_cv]  # get curvature of first and last lens interfaces

            # check which lens side is the more curved.
            # "first_curv" = first surface is more curved than last one, "last_curv" = last surface more curved
            # "same" = radius of curvature of both extremum sides of the lens are the same
            if np.abs(self.curv[0]) > np.abs(self.curv[1]):
                self.curv_state = "first_curv"
            elif np.abs(self.curv[0]) < np.abs(self.curv[1]):
                self.curv_state = "last_curv"
            else:  # curvature before and after lens are same
                self.curv_state = "same"

        else:  # filepath is None
            self.efl = None
            self.curv_state = None
            self.diameter_minmax = self.curv = [None, None]
            self.opm = None

    def check_for_air(self, suppress_air_gap=False, idx=None):
        # check if there is air gap before or after the lens
        # can suppress air gap if required (but will delete diaphragm if the air gap exist for a good reason)
        # idx => lens interface id [first interface, last interface]
        if idx is None:
            # idx=self.idx
            idx = [1, self.sm.get_num_surfaces() - 2]

        i = idx[0]
        idx_no_air = idx.copy()
        air_gap = False
        layers_to_delete_before = []
        layers_to_delete_after = []

        # look for air part BEFORE the lens
        # so from the first interface after the object, to the last lens interface
        # look all index from first surface to last one until a layer is not air (has not a refractive index of 1
        # (for tested wavelengths : the ones of the system))
        while i < idx[1] and (np.array(self.sm.rndx[i]) == 1.0).all():
            if suppress_air_gap:
                layers_to_delete_before.append(i)
            air_gap = True
            idx_no_air[0] = i
            i = i + 1

        # look for air part AFTER the lens
        # so from the last interface before the image, to the first lens interface
        j = j0 = idx_no_air[1]

        while j > idx_no_air[0] and (np.array(self.sm.rndx[j]) == 1.0).all():
            if suppress_air_gap:
                layers_to_delete_after.append(j)
            if j != j0:
                # if just before image is the only air_gap, it is not considered as an airgap,
                # as it is necessary to have a last lens interface curvature)
                # (but will still remove air distance with image if suppress_air_gap is True)
                air_gap = True
            idx_no_air[1] = j
            j = j - 1

        if suppress_air_gap:
            if len(layers_to_delete_after) > 0:  # if is not empty
                # to not confuse index, start by the last one to the first one
                for i in range(len(layers_to_delete_after) - 1, 0, -1):
                    self.sm.set_cur_surface(layers_to_delete_after[i])
                    self.sm.remove()
                self.sm.gaps[layers_to_delete_after[0]].thi = 0  # distance to the image !

            for i in range(len(layers_to_delete_before) - 1, -1, -1):
                self.sm.set_cur_surface(layers_to_delete_before[i])
                self.sm.remove()

            self.opm.update_model()
            idx_no_air = [1, self.sm.get_num_surfaces() - 2]

        self.idx = idx_no_air
        return air_gap

    def get_diameter(self):
        # this function return the min and max semi-diameters of all surfaces of the lens

        # get an array with all surface semi diamter (except object and image)
        sd_np = np.ones(self.sm.get_num_surfaces())  # preallocating

        for i, ifcs in enumerate(self.sm.ifcs):
            sd = ifcs.surface_od()  # get semi diameter of surface index i
            if sd is not None:
                sd_np[i] = sd
            else:
                sd_np[i] = np.nan

        # get min and max semi diameter of all surfaces of the lens
        sd_min = np.nanmin(sd_np[1:len(sd_np) - 1])
        sd_max = np.nanmax(sd_np[1:len(sd_np) - 1])

        # diameter = 2*semi-diameter
        dia_min = sd_min * 2
        dia_max = sd_max * 2

        return [dia_min, dia_max]

    def flip(self):
        # this function flip the orientation of the lens along the optical axis

        # flip opm model
        self.opm.flip()
        self.opm.update_model()

        # flip orientation dependant parameters
        # surface index of the lens
        idxtps = self.idx.copy()
        self.idx[0] = self.sm.get_num_surfaces() - 1 - idxtps[1]
        self.idx[1] = self.sm.get_num_surfaces() - 1 - idxtps[0]

        # curvature of lens
        self.curv = [self.curv[1], self.curv[0]]

        if self.curv_state == "first_curv":
            self.curv_state = "last_curv"
        elif self.curv_state == "last_curv":
            self.curv_state = "first_curv"

    def schematic_raytracing_plt(self, do_draw_rays=True, do_paraxial_layout=False):
        # Draw ray trace for the user
        layout_plt = plt.figure(FigureClass=InteractiveLayout, opt_model=self.opm,
                                do_draw_rays=do_draw_rays, do_paraxial_layout=do_paraxial_layout).plot()
        return layout_plt



import logging
import time
import psutil
import multiprocess as mp

from pyrftl.func.various_stuff import set_log


def log_process(queue_log, param):
    # process to create logs from the log queue. Process stop when queue_log.get() return None

    logger = logging.getLogger('PyRFTL')

    set_log(param['save_folder_path'], False)

    while True :
        # if there is an element in log queue, create the log, if not wait and check again. When element None is in the
        # queue, that mean the end of logging messages so the process should stop
        if not queue_log.empty() :
            element = queue_log.get()

            if element is None :
                # Stop
                break
            else:
                try :
                    [text, level] = element
                    logger.log(level, text)

                except Exception as exception:
                    logger.error("Error while logging, unable to create the log.\n"
                                 + "Error is :\n"
                                 + str(exception) + "\n")
        else :
            time.sleep(0.05)


def init_global_pair_id (count) :
    global pair_id_
    pair_id_ = count


def create_proc_idx_queue (param) :
    if param['force_affinity_core']:  # if one want to force each subprocess to execute on a separated core
        # create a multiprocessing queue with core index to execute for each process
        q = mp.Queue()
        nb_cores = psutil.cpu_count()
        for i_proc in range(param['nb_subprocesses']):
            i_proc_mod = i_proc % nb_cores  # get the modulo to have always a number between 0 and nb_cores-1
            core_id = psutil.Process().cpu_affinity()[i_proc_mod]
            q.put(core_id)
    else:
        q = None

    return q



import customtkinter as ctk
from pyrftl.gui.helpbox import HelpWindowMain
import tkinter


class Frame(ctk.CTkFrame):
    # frame with a textbox which show some details of the optical model, and allow to modify it
    def __init__(self, master, new_frame=False, help_window=None, **kwargs):
        # initialize the object
        super().__init__(master, **kwargs)

        if new_frame:
            self.frame = ctk.CTkFrame(master)
            self.frame.pack(fill=tkinter.BOTH, expand=True)

        else:
            self.frame = master

        if hasattr(master, 'font'):
            self.font = master.font
            self.font.weight = "normal"
        else:
            self.font = ctk.CTkFont(size=15, weight="normal")

        if help_window is not None:
            self.help_window = help_window
        else:
            self.help_window = HelpWindowMain(font)



import logging
import sys  # system, to set logging to file and stdout
import os.path

# import from PyRFTL
logger = logging.getLogger('PyRFTL')



#### LOG ####
def set_log(folder_name=None,log_stdout=True) :
    # if folder_name exist, a log file would be put in the folder
    # if log_stdout is True, log would also be output to stdout
    # it should not be folder_name=None and log_stdout=True at the same time. (if it is, standard basic config would be used)

    handlers = []
    if folder_name is not None :
        handlers.append(logging.FileHandler(os.path.join(folder_name, "log.txt")))
    if log_stdout :
        handlers.append(logging.StreamHandler(sys.stdout))

    logging.basicConfig(handlers=handlers, encoding='utf-8', level=logging.INFO)


#### DISTANCES ####
def lens_distance_to_efl(distance, opm, gap_index):
    # change the distance between the two lenses and return the new effective focal length
    opm['seq_model'].gaps[gap_index].thi = distance
    opm.update_model()
    efl = opm['parax_model'].opt_model['analysis_results']['parax_data'].fod.efl
    return efl


def safe_lens_distance_to_efl(distance, opm, gap_index):
    try:
        efl = lens_distance_to_efl(distance, opm, gap_index)
    except:
        efl = None
    return efl


def lens_distance_to_efl_zero(distance, opm, gap_index, expected_focal_length):
    # to be used by obtainDistanceToCrossAZero for optimization
    return lens_distance_to_efl(distance, opm, gap_index) - expected_focal_length


def obtain_distance_to_cross_a_zero(opm, gap_index, expected_fl, actual_dist):
    # search for an interval [a,b] of distances between the two lenses, for which the expected efl for the pair is between the pair efl when distance between lenses is a and when distance is b
    # actual_dist is the initialize distance, which is expected to be the optimal distance. It should be >= 0 (physically impossible if <0)
    if actual_dist < 0:
        raise ValueError("initialize distance of obtainDistanceToCrossAZero cannot be < 0")

    epsilon = 0.5
    a = actual_dist - epsilon
    b = actual_dist

    if a < 0:
        a = 0

    nbboucle = 0
    efl_zero_for_separation_a = lens_distance_to_efl_zero(a, opm, gap_index, expected_fl)
    efl_zero_for_separation_b = lens_distance_to_efl_zero(b, opm, gap_index, expected_fl)

    should_continue = efl_zero_for_separation_a * efl_zero_for_separation_b > 0

    while should_continue:
        # change distance b
        b = actual_dist + epsilon
        efl_zero_for_separation_b = lens_distance_to_efl_zero(b, opm, gap_index,
                                                              expected_fl)  # determine the efl (rather efl - expected_efl) of the pair when both lenses are separated of distance a
        should_continue = efl_zero_for_separation_a * efl_zero_for_separation_b > 0  # check if expected_efl is between efl_with_distance_a and efl_with_distance_b

        epsilon = epsilon * 2  # have already changed a, so epsilon change between change of a and b

        # check the loop number, and increase it
        if nbboucle > 8:  # avoid infinity loop. At loop 8, a is initial-128 and b is initial+128 !
            logger.error(
                'impossible to define an interval of lenses separation distances that permit to find the effective focal length')
            logger.error("a " + str(a) + " " + str(lens_distance_to_efl_zero(a, opm, gap_index, expected_fl)) + "\nb " + str(
                b) + " " + str(lens_distance_to_efl_zero(b, opm, gap_index, expected_fl)))
            should_continue = False
        else:
            nbboucle = nbboucle + 1

        # change distance a
        if a > 0 and should_continue:
            a = actual_dist - epsilon  # change distance a
            if a < 0:
                a = 0  # distance between lenses cannot be < 0
            efl_zero_for_separation_a = lens_distance_to_efl_zero(a, opm, gap_index,
                                                                  expected_fl)  # determine the efl (rather efl - expected_efl) of the pair when both lenses are separated of distance a
            should_continue = efl_zero_for_separation_a * efl_zero_for_separation_b > 0  # check if expected_efl is between efl_with_distance_a and efl_with_distance_b

    return [a, b, nbboucle]


#### FLOAT COMPARISON ####

def is_between(number_to_test, inf=None, sup=None):
    # said if inf <= number_to_test <= sup
    # inf and sup can be None (so no inf and no sup values)
    if (not inf is None) and number_to_test < inf:
        return False

    if (not sup is None) and number_to_test > sup:
        return False

    return True


#### OPTICAL MODEL ####

def opm_dist_idx1_idx2(opm, idx1, idx2):
    # return the distance between idx1 and idx2 in optical model opm
    # gaps of layer idx2 is not get, as it is between idx2 and idx2+1
    dist = 0
    for i in range(idx1, idx2):
        dist = dist + opm['seq_model'].gaps[i].thi
    return dist


#### OTHERS ####

def time_str_hms_from_s(t_total):
    hours = t_total // 3600
    minutes = (t_total - hours * 3600) // 60
    seconds = round(t_total - hours * 3600 - minutes * 60)
    return str(hours) + "h " + str(minutes) + "min " + str(seconds) + 's'


def check_param(param):
    if param['thin_lens_first_analysis'] and (not is_between(param['tolerance_prcent'], 0, 100)):
        raise ValueError('tolerance_prcent should be between 0 and 100 (or None)')

    if not isinstance(param['best_focus_fast'], bool):
        raise ValueError('best_focus_fast should be a boolean')

    if not param['cut_off_angle_min'] < param['cut_off_angle_test_end']:
        raise ValueError('cut_off_angle_min > cut_off_angle_test_end, it should be the opposite')

    if not (param['wavefront_analysis'] == "fast" or param['wavefront_analysis'] == "mean"):
        raise ValueError('wavefront_analysis should be "fast" or "mean"')


def description_pair_list(param, pair={}, columns_names=True, first_rms=None):
    # output some data of the pair in a list
    # header specify if output column names also or only the data
    # put pair = None if wants only the column names

    description = []
    if not isinstance(pair, dict):
        pair = {'pair':pair}

    # create column names
    if columns_names:
        columns_names_list = ['lens1 filename', 'lens2 filename', 'config', 'pair short name',
                              'lens1 id', 'lens2 id', 'lens1 EFL', 'lens2 EFL',
                              'first analysis RMS wavefront error',
                              'second analysis RMS wavefront error', '<lambda/14', 'cut-off angle', 'ffl',
                              'distance between lenses',
                              'bfl', 'optimum bfl for RMS wavefront', 'efl', "distance F to F'",
                              "distance first to last surface of pair lenses",
                              "lens1 total path", "lens2 total path", 'lens1 min diameter', 'lens1 max diameter',
                              'lens2 min diameter', 'lens2 max diameter',
                              'lens1 info', 'lens2 info']

        nb_wvl = len(param['wvl_and_weight'])
        nb_fld = 1
        new_columns = ['fld', 'wvl', 'RMS'] * nb_fld * nb_wvl
        columns_names_list = columns_names_list + new_columns  # concatenate

        description.append(columns_names_list)

    # extract the data
    for pair_key in pair :
        pair_tmp = pair[pair_key]

        # if want anonymity of lenses path or not
        if param['output_csv']['hide_path_in_csv'] :
            tmp_path_lens1 = os.path.relpath(pair_tmp.lens1.filepath, param['lenses_param']['LensesFolderPath'])
            tmp_path_lens2 = os.path.relpath(pair_tmp.lens2.filepath, param['lenses_param']['LensesFolderPath'])
        else:
            tmp_path_lens1 = pair_tmp.lens1.filepath
            tmp_path_lens2 = pair_tmp.lens2.filepath

        tmp_filename_lens1 = os.path.basename(pair_tmp.lens1.filepath)
        tmp_filename_lens2 = os.path.basename(pair_tmp.lens2.filepath)

        if first_rms is None:
            rms1 = pair_tmp.rms_detail[0]['RMS']
            rms2 = None
            lambda14 = rms1 < 1/14
        else:
            rms1 = first_rms
            rms2 = pair_tmp.rms_detail[0]['RMS']
            lambda14 = rms2 < 1 / 14

        if hasattr(pair_tmp, 'foc_opti') and pair_tmp.foc_opti.success:
            bfl_opti = pair_tmp.bfl + pair_tmp.foc_opti.x[0]
        else:
            bfl_opti = None

        dist_first_last_surface_lens = opm_dist_idx1_idx2(pair_tmp.opm, pair_tmp.lens1surf[0], pair_tmp.lens2surf[1])
        dist_FFi = dist_first_last_surface_lens - pair_tmp.ffl + pair_tmp.bfl  # distance from object focal plane to image focal plane
        efl = pair_tmp.pm.opt_model['analysis_results']['parax_data'].fod.efl

        data_list = [tmp_filename_lens1, tmp_filename_lens2,
                     str(pair_tmp.config), pair_tmp.short_name, str(pair_tmp.lens1.id), str(pair_tmp.lens2.id),
                     pair_tmp.lens1.efl, pair_tmp.lens2.efl, rms1,
                     rms2, lambda14, pair_tmp.cutoffangle['value'], pair_tmp.ffl, pair_tmp.lenses_gap,
                     pair_tmp.bfl, bfl_opti, efl, dist_FFi, dist_first_last_surface_lens,
                     tmp_path_lens1, tmp_path_lens2,
                     pair_tmp.lens1.diameter_minmax[0], pair_tmp.lens1.diameter_minmax[1],
                     pair_tmp.lens2.diameter_minmax[0], pair_tmp.lens2.diameter_minmax[1],
                     pair_tmp.lens1.info, pair_tmp.lens2.info]

        try :
            if hasattr(pair_tmp, 'wvf_preprocessed'):
                if pair_tmp.wvf_preprocessed :  # not empty
                    nb_wvl = len(pair_tmp.wvf_preprocessed[0]['wvl_and_data'])

                    for i_fld in range(len(pair_tmp.wvf_preprocessed)):  # fld
                        for i_wvl in range(nb_wvl):  # wvl
                            tmp_fld = pair_tmp.rms_detail[i_fld * nb_wvl + i_wvl + 1]['used_fld']
                            tmp_wvl = pair_tmp.rms_detail[i_fld * nb_wvl + i_wvl + 1]['used_wvl']
                            tmp_rms = pair_tmp.rms_detail[i_fld * nb_wvl + i_wvl + 1]['RMS']
                            data_list = data_list + [tmp_fld, tmp_wvl, tmp_rms]
        except:
            pass

        description.append(data_list)

    return description


def pair_df_2_lens_config(pair_dataframe, linekey, lenses_list):
    # export config and original lenses from a pair
    config = pair_dataframe.at[linekey, 'config']
    if str(config) in ['nan', "None", 'none']:
        config = None

    lens1 = lenses_list[int(pair_dataframe.at[linekey, 'lens1 id'])]

    lens2_id = pair_dataframe.at[linekey, 'lens2 id']
    if str(lens2_id) in ['nan', "None", 'none']:
        lens2 = None
    else:
        lens2 = lenses_list[int(lens2_id)]

    return config, lens1, lens2




import tkinter
import customtkinter as ctk
import numpy as np

# from rayoptics.environment import FieldSpec
import matplotlib.pyplot as plt

from pyrftl.gui.helpbox import HelpWindowMain, create_help_box
from pyrftl.func.save import sort_pair_rms, detail_seq_model_str
from pyrftl.func.pairs import Pair
from pyrftl.func.various_stuff import pair_df_2_lens_config
from pyrftl.gui.wavefront_box import WavefrontConfigMain


class PairDfDetail(ctk.CTkFrame):
    def __init__(self, master, pair_dataframe, pair_dic, lenses_list, param=None,
                 help_window=None, **kwargs):
        # initialize the object
        super().__init__(master, **kwargs)

        # delete previous frame
        if hasattr(master, 'main_frame'):
            master.main_frame.destroy()

        # create a main frame
        master.main_frame = ctk.CTkFrame(master, fg_color="transparent")
        master.main_frame.pack(fill=tkinter.BOTH, expand=True)

        self.main_frame = master.main_frame

        if hasattr(master, 'font'):
            font = master.font
        else:
            font = ctk.CTkFont(size=15, weight="bold")

        if help_window is not None:
            self.help_window = help_window
        else:
            self.help_window = HelpWindowMain(font)

        self.pair_df = pair_dataframe
        self.pair_dic = pair_dic
        self.lenses_list = lenses_list
        self.param = param

        if param is not None :
            try:
                self.label_result_directory = ctk.CTkLabel(self.main_frame, font=font,
                                                           text= "Results are exported in : \n" + param['save_folder_path'])
                self.label_result_directory.grid(row=0, column=0, padx=5, pady=1, columnspan=3, sticky="w")
            except Exception as exception:
                pass

        self.textbox_pair_short_info = ctk.CTkTextbox(self.main_frame, wrap="word", font=font, width=500,
                                                      height=80)
        self.textbox_pair_short_info.grid(row=1, column=0, padx=5, pady=(0,10), columnspan=2, sticky="w")
        self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only

        self.button_change_pair = ctk.CTkButton(self.main_frame, text="Change pair", command=self.change_pair_menu,
                                                font=font)
        self.button_change_pair.grid(row=1, column=2, padx=10, pady=(10,30), sticky='n')

        txt_nopair = ('\n\n'
                      'How to deal with it:\n'
                      '- Remove distance constraints\n'
                      '- Add more lenses to import\n'
                      '- Reduce the spectral range\n'
                      '- Increase the required focal length (for remote focusing, this also implies modifying the '
                      'other tube lens)\n'
                      '- Disable the thin lens filtering (if it was enabled)')

        if pair_dataframe is None :
            self.button_change_pair.configure(state="disabled")
            self.textbox_pair_short_info.configure(state="normal")  # configure textbox to be modifiable
            self.textbox_pair_short_info.insert("0.0", 'No pair has been selected !\n'
                                                       'Please change your criteria'+txt_nopair)  # add text
            self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only

        else:
            # some pairs have been selected. Detect the best one and display it
            self.pair_detail_frame = ctk.CTkFrame(self.main_frame)
            self.pair_detail_frame.grid(row=2, column=0, columnspan=3)
            self.pair_detail_obj = PairDetail(self.pair_detail_frame, param=param, help_window=help_window)

            # get best pair
            if self.pair_df.iloc[0].at['<lambda/14'] :  # better rms than diffraction
                self.change_pair(self.pair_df.index[0])
                self.best_pair_sn = self.pair_df.index[0]
            else :  # lower rms than diffraction
                self.best_pair_sn = None
                self.textbox_pair_short_info.configure(state="normal")  # configure textbox to be modifiable
                self.textbox_pair_short_info.insert("0.0", 'No pair has a better rms than 1/14 !\n'
                                                           'Please change your criteria'+txt_nopair)  # add text
                self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only

    def change_pair_menu(self):
        dialog_box = ctk.CTkInputDialog(text="Enter pair short name:", title="Change displayed pair")
        pair_sn = dialog_box.get_input()

        if pair_sn in self.pair_df.index :
            self.change_pair(pair_sn)
        else:
            txt = ('Please select a correct pair short name\n'
                   'see csv file for the list of them. Please respect letter case.'
                   '   best (if exist) is : ' + str(self.best_pair_sn))
            self.change_txt_short_info_box(txt)

            self.pair_detail_obj.pair = None
            self.pair_detail_obj.update_textbox()

    def change_pair(self, pair_short_name):
        # change the displayed pair in PairDetail
        if pair_short_name not in self.pair_df.index :
            txt = 'Error, no pair named : ' + pair_short_name
            self.change_txt_short_info_box(txt)
            self.pair_detail_obj.pair = None
            self.pair_detail_obj.update_textbox()

        else:
            # change information in the short info box
            txt = ('pair : ' + pair_short_name + '  : ' +
                   self.pair_df.at[pair_short_name, 'lens1 filename'] + ' --> ' +
                   self.pair_df.at[pair_short_name, 'lens2 filename'] + ' config ' +
                   self.pair_df.at[pair_short_name, 'config'])
            txt = txt + '\ncut-off angle : ' + str(self.pair_df.at[pair_short_name, 'cut-off angle']) + '°   '

            if self.pair_df.at[pair_short_name, 'second analysis RMS wavefront error'] is None:
                rms = self.pair_df.at[pair_short_name, 'first analysis RMS wavefront error']
                grid = str(self.param['nbRays_first_comparison'])
            else:
                rms = self.pair_df.at[pair_short_name, 'second analysis RMS wavefront error']
                grid = str(self.param['nbRays_high_comparison'])
            txt = txt + 'rms grid ' + grid + ' : ' + f'{rms:.3g}' + ' waves'

            self.change_txt_short_info_box(txt)

            # change the displayed pair in pair_detail_obj (big info box + response to buttons with ray tracing etc)
            if hasattr(self.pair_dic, pair_short_name):
                # pair already exist in the dictionary
                self.pair_detail_obj.pair = pair = self.pair_dic[pair_short_name]
            else:
                # pair not in the dictionary, need to (re)create it from the dataframe indications (to have more info
                # than just the ones in the dataframe, else can just use the csv file if dataframe infos are enough...)
                self.pair_detail_obj.update_textbox("Computing...")

                # took infos from the dataframe to be able to create the pair
                config, lens1, lens2 = pair_df_2_lens_config(self.pair_df, pair_short_name, self.lenses_list)

                # create the pair
                pair = Pair(lens1, lens2, self.param['expected_focal_length'], None,
                            {'lens1 id :', lens1.id, 'lens2 id :', lens2.id})
                pair.create_optical_model(config=config, wvl_and_weight=self.param['wvl_and_weight'],
                                          pupil=self.param['pupil'], fld=[0],
                                          dist_pupil_pair=self.param['dist_pupil_pair'])

                self.pair_detail_obj.pair = self.pair_dic[pair_short_name] = pair
            self.pair_detail_obj.update_textbox()

    def change_txt_short_info_box(self, txt):
        self.textbox_pair_short_info.configure(state="normal")  # configure textbox to be modifiable
        self.textbox_pair_short_info.delete("0.0", "end")
        self.textbox_pair_short_info.insert("0.0", txt)  # add text
        self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only


class PairOneDetail(ctk.CTkFrame):
    def __init__(self, master, pair, param_one, help_window=None, **kwargs):
        # initialize the object
        super().__init__(master, **kwargs)

        # delete previous frame
        if hasattr(master, 'main_frame'):
            master.main_frame.destroy()

        # create a main frame
        master.main_frame = ctk.CTkFrame(master, fg_color="transparent")
        master.main_frame.pack(fill=tkinter.BOTH, expand=True)

        self.main_frame = master.main_frame

        if hasattr(master, 'font'):
            font = master.font
        else:
            font = ctk.CTkFont(size=15, weight="bold")

        if help_window is not None:
            self.help_window = help_window
        else:
            self.help_window = HelpWindowMain(font)

        self.pair = pair
        rms = pair.rms_detail[0]['RMS']
        grid = pair.rms_detail[0]['grid size']
        cut_off = pair.cutoffangle['value']
        self.param_one = param_one

        self.textbox_pair_short_info = ctk.CTkTextbox(self.main_frame, wrap="word", font=font, width=700,
                                                      height=80)
        self.textbox_pair_short_info.grid(row=1, column=0, padx=5, pady=(0,10), columnspan=2, sticky="w")
        txt = ("pair : " + pair.name() +
               "\nCut-Off angle [°]: " + str(cut_off) + "\nRMS grid " + str(grid) + " : " +
               "{0:.3g}".format(rms) + " waves")

        self.change_txt_short_info_box(txt)

        self.pair_detail_frame = ctk.CTkFrame(self.main_frame)
        self.pair_detail_frame.grid(row=2, column=0, columnspan=3)
        self.pair_detail_obj = PairDetail(self.pair_detail_frame, pair=pair, param=param_one, help_window=help_window)

    def change_txt_short_info_box(self, txt):
        self.textbox_pair_short_info.configure(state="normal")  # configure textbox to be modifiable
        self.textbox_pair_short_info.delete("0.0", "end")
        self.textbox_pair_short_info.insert("0.0", txt)  # add text
        self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only


class PairDetail(ctk.CTkFrame):
    # frame with a textbox which show some details of the pair, and with buttons to disp the wavefront, the ray diagram,
    # to save it etc
    def __init__(self, master, pair=None, new_frame=False, param=None, help_window=None, **kwargs):
        # initialize the object
        super().__init__(master, **kwargs)

        if new_frame:
            self.frame = ctk.CTkFrame(master)
            self.frame.pack(fill=tkinter.BOTH, expand=True)

        else:
            self.frame = master

        if hasattr(master, 'font'):
            font = master.font
            font.weight = "normal"
        else:
            font = ctk.CTkFont(size=15, weight="normal")

        self.font=font

        if help_window is not None:
            self.help_window = help_window
        else:
            self.help_window = HelpWindowMain(font)

        self.wavefront_box = WavefrontConfigMain(pair_detail_obj=self, font=font)

        self.label_details_text = ctk.CTkLabel(self.frame, font=font, text="Pair details :")
        self.label_details_text.grid(row=0, column=0, padx=5, pady=1, columnspan=2, sticky="w")

        self.textbox_details = ctk.CTkTextbox(self.frame, wrap="word", width=700, height=400,
                                              font=ctk.CTkFont(family="Lucida Console", size=14))
                                              # font with same space for all characters
        self.textbox_details.grid(row=0, column=0, padx=5, pady=(0,10), columnspan=5, sticky="w")
        self.textbox_details.configure(state="disabled")  # configure textbox to be read-only

        self.pair = pair
        self.param = param

        self.update_textbox()

        self.button_ray = ctk.CTkButton(self.frame, text="Ray diagram", command=self.raydiagram, font=font)
        self.button_ray.grid(row=1, column=0, padx=10, pady=(10,30), sticky='n')

        self.button_wvf = ctk.CTkButton(self.frame, text="Wavefront", command=self.wavefront, font=font)
        self.button_wvf.grid(row=1, column=1, padx=10, pady=(10,30), sticky='n')

        self.button_wvf = ctk.CTkButton(self.frame, text="Save pair .roa", command=self.save_roa, font=font)
        self.button_wvf.grid(row=1, column=2, padx=10, pady=(10,30), sticky='n')

    def update_textbox(self, txt=None):
        if type(txt) is not str :
            if self.pair is None :
                detail_txt = ('No pair selected, nothing to display !\n'
                              'Please use the "Change pair" button to select a pair.')
            else:
                detail_txt = detail_seq_model_str(self.pair)
        else:
            detail_txt = txt

        self.textbox_details.configure(state="normal")  # configure textbox to be modifiable
        self.textbox_details.delete("0.0", "end")  # delete all text
        self.textbox_details.insert("0.0", detail_txt)  # add text
        self.textbox_details.configure(state="disabled")  # configure textbox to be read-only

    def change_pair(self, new_pair):
        self.pair = new_pair
        self.update_textbox()

    def raydiagram(self):
        # display the pair ray tracing diagram
        if self.pair is not None:
            try :
                dialog_box = ctk.CTkInputDialog(text="Enter field angle (°):",
                                                title="Ray diagram")
                angle = dialog_box.get_input()

                try:
                    angle_float = float(angle)
                except:
                    if self.param is not None:
                        angle_float = self.param['cut_off_angle_min']
                    else:
                        angle_float = 0

                self.pair.change_field_angle(angle_float)

                self.pair.disp_raytracing()
            except Exception as exception:
                print('Impossible to display the ray diagram because of this error :\n'
                      + str(exception))

    def wavefront(self):
        if self.pair is not None:
            try :
                self.wavefront_box.create_wavefrontconfig(font=self.font)
            except Exception as exception :
                print('oups')
                print(exception)

    def save_roa(self):
        try:
            filepath = tkinter.filedialog.asksaveasfilename(defaultextension='.roa')
            self.pair.opm.save_model(filepath)
        except:
            pass



